import json
import importlib
import copy

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('AppIndicator3', '0.1')
gi.require_version('Notify', '0.7')

from gi.repository import Notify as notify
from gi.repository import Gtk, GLib, Gdk

import conf 

import utils 
from utils import *

def update_user_settings_value(widget = None, key_val = None):
    print("update user_settings")
    print(key_val)

    # TODO: make this work with deeper values,         
    # accept key1, key2, key3

    if key_val:
        conf.user[key_val['key']] = key_val['value']

        save_user_settings()


def update_connector_setting(widget, app, connector_category, id, key, widget_getter = "Add this feature! :)"):
    conf.user[connector_category][id][key] = widget.get_active()
    save_user_settings()
    app.async_refresh()


class SettingsWindow(Gtk.Window):

    def __init__(self, parent=None, **kwargs):
        self.app = parent
        Gtk.Window.__init__(self, title="Settings")
        self.set_border_width(15)
        self.set_position(position=1) # Works on x11 but not wayland (ubuntu 22.04)

        self.connect("destroy", self.on_close)

        row = 0
        self.notebook = Gtk.Notebook()
        self.add(self.notebook)

        grid = Gtk.Grid()
        grid.set_row_spacing(10)
        grid.set_column_spacing(10)

        self.notebook.append_page(grid,Gtk.Label('Todolists and Time Trackers'))


        sub_head = Gtk.Label()
        sub_head.set_markup("<b>Todo Lists</b>")
        grid.attach(sub_head,0,(row:=row+1),5,1)

        self.boxes = {}
        self.inputs = {}

        self.boxes['todolists_outer'] = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        grid.attach(self.boxes['todolists_outer'],0,(row:=row+1),5,1)

        # Gets replaced when lists are loaded
        self.boxes['todolists_inner'] = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        self.boxes['todolists_outer'].add(self.boxes['todolists_inner'])

        separator = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
        self.boxes['todolists_outer'].pack_start(separator, True, True, 0)
        
        self.show_connectors('todolists')


        grid.attach(Gtk.Label(label="Connect a new Todo List"),0,(row:=row+1),2,1)  
        self.inputs['new_todolists_type'] = Gtk.ComboBoxText()

        for item in conf.connectors['todolists']:
            self.inputs['new_todolists_type'].append(item,item)

        self.inputs['new_todolists_type'].set_active_id(list(conf.connectors['todolists'].keys())[0])
        grid.attach(self.inputs['new_todolists_type'],2,row,2,1)

        button = Gtk.Button(label="Go")
        button.connect("clicked", self.edit_add_connector,'todolists')
        grid.attach(button,4,row,1,1)



        grid.attach(Gtk.Box(border_width=10),0,(row:=row+1),5,1) # Spacer

        sub_head = Gtk.Label()
        sub_head.set_markup("<b>Time Trackers</b>")
        grid.attach(sub_head,0,(row:=row+1),5,1)

        self.boxes['timetrackers_outer'] = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        grid.attach(self.boxes['timetrackers_outer'],0,(row:=row+1),5,1)


        # Gets replaced when lists ar loaded
        self.boxes['timetrackers_inner'] = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        self.boxes['timetrackers_outer'].add(self.boxes['timetrackers_inner'])

        separator = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
        self.boxes['timetrackers_outer'].pack_start(separator, True, True, 0)
        
        self.show_connectors('timetrackers')


        grid.attach(Gtk.Label(label="Connect a Time Tracker"),0,(row:=row+1),2,1)  
        self.inputs['new_timetrackers_type'] = Gtk.ComboBoxText()

        for item in conf.connectors['timetrackers']:
            self.inputs['new_timetrackers_type'].append(item,item)

        self.inputs['new_timetrackers_type'].set_active_id(list(conf.connectors['timetrackers'].keys())[0])
        grid.attach(self.inputs['new_timetrackers_type'],2,row,2,1)


        button = Gtk.Button(label="Add")
        button.connect("clicked", self.edit_add_connector,'timetrackers')
        grid.attach(button,4,row,1,1)
        




        # Settings
        row = 0
        grid = Gtk.Grid()
        grid.set_row_spacing(10)
        grid.set_column_spacing(10)
        self.notebook.append_page(grid,Gtk.Label('Settings'))


        grid.attach(Gtk.Box(border_width=10),0,(row:=row+1),5,1) # Spacer

        grid.attach(self.settings_updater('default_text','What am i Doing?','Entry'),0,(row:=row+1),5,1)

        grid.attach(self.settings_updater('pomodoro_interval',25,'SpinButton',"(minutes)"),0,(row:=row+1),5,1)

        grid.attach(self.settings_updater('randomness_interrupt_interval', 3, 'SpinButton',"\nPop up todo menu every so many minutes if not doing a task"),0,(row:=row+1),5,1)

        grid.attach(self.settings_updater('todolist_refresh_interval', 3, 'SpinButton',"(hours)"),0,(row:=row+1),5,1)
        
        grid.attach(self.settings_updater('max_top_level_menu_items', 18, 'SpinButton',""),0,(row:=row+1),5,1)

        grid.attach(self.settings_updater('tick_interval', 18, 'SpinButton',"\nMust be an even number fraction of 60. (requires restart)"),0,(row:=row+1),5,1)
        
        grid.attach(self.settings_updater('display_todolist_as_top_level_list', 'auto', 'ComboBoxText',"", options = ['auto','always','never']),0,(row:=row+1),5,1)
        
        grid.attach(self.settings_updater('prompts', None, 'TextView',""),0,(row:=row+1),5,1)

 


        # Time Targets notebook page  
        self.targets_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        self.notebook.append_page(self.targets_box,Gtk.Label('Time Targets'))

        self.boxes['lists_time_targets_outer'] = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        self.targets_box.add(self.boxes['lists_time_targets_outer'])
        # Gets replaced when targets are loaded
        self.boxes['lists_time_targets_inner'] = Gtk.Box()
        self.boxes['lists_time_targets_outer'].add(self.boxes['lists_time_targets_inner'])

        separator = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
        self.boxes['lists_time_targets_outer'].pack_start(separator, True, True, 0)
        
        self.show_time_targets('lists')

        add_lists_target_button = Gtk.Button(label="Add List Time Target")
        add_lists_target_button.connect("clicked", self.edit_add_target,'lists')

        self.targets_box.add(add_lists_target_button)


        self.boxes['tasks_time_targets_outer'] = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        self.targets_box.add(self.boxes['tasks_time_targets_outer'])
        # Gets replaced when targets are loaded
        self.boxes['tasks_time_targets_inner'] = Gtk.Box()
        self.boxes['tasks_time_targets_outer'].add(self.boxes['tasks_time_targets_inner'])

        separator = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
        self.boxes['tasks_time_targets_outer'].pack_start(separator, True, True, 0)
        
        self.show_time_targets('tasks')

        add_tasks_target_button = Gtk.Button(label="Add Task Time Target")
        add_tasks_target_button.connect("clicked", self.edit_add_target,'tasks')

        self.targets_box.add(add_tasks_target_button)

        self.show_all()



    def settings_updater(self, label, default_value = '', input_type = 'Entry', notes = '', options = False):
        # returns a box widget with a label and a input that live-updates user settings 
        
        if label in conf.user:
            value = conf.user[label]
        else: 
            value = default_value

        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=5)

        label_widget = Gtk.Label()
        label_widget.set_markup("<b>"+label.replace('_', ' ').title()+"</b> "+str(notes))
        box.pack_start(label_widget,False,False,5)

        match (input_type):
            case ("SpinButton"):
                adjustment = Gtk.Adjustment(value=int(value),
                            lower=0,
                            upper=999,
                            step_increment=1,
                            page_increment=5,
                            page_size=0)

                value_widget = Gtk.SpinButton(adjustment=adjustment,climb_rate=1,digits=1)

                value_widget.connect(
                    "value-changed", 
                    lambda value_widget : update_user_settings_value(None, {'key':label,'value':value_widget.get_value_as_int()})
                    )

            case ("Entry"):
                value_widget = Gtk.Entry() 
                value_widget.set_text(str(value))
                value_widget.connect("changed",
                    lambda value_widget: update_user_settings_value(None, {'key':label,'value':value_widget.get_text()})
                    ) 

                    
            case ("TextView"):
                buffer = Gtk.TextBuffer()
                buffer.set_text(str(value))

                value_widget = Gtk.TextView(buffer=buffer) 
                buffer.connect("changed",
                    lambda value_widget: update_user_settings_value(None, {'key':label,'value':buffer.get_text(buffer.get_bounds()[0], buffer.get_bounds()[1], False)})
                    )
                    
            case ("ComboBoxText"):

                value_widget = Gtk.ComboBoxText() 
                for item in options:
                    value_widget.append(item,item)
                value_widget.set_active_id(str(value))
                value_widget.connect("changed",
                    lambda value_widget: update_user_settings_value(None, {'key':label,'value':value_widget.get_active_text()})
                    )

        box.pack_end(value_widget,False,False,5)
        
        return box

    def show_time_targets(self,category="tasks"):
        outer_box = self.boxes[category+"_time_targets_outer"]

        self.boxes[category+"_time_targets_inner"].destroy()

        self.boxes[category+"_time_targets_inner"] = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        inner_box = self.boxes[category+"_time_targets_inner"]
        outer_box.add(inner_box)

        for id, data in conf.user['time_targets'][category].items():

            box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
            label = Gtk.Label()
            
            item_data = utils.db_get_item_by_id(id,category)
            if not item_data:
                # error_notice(title="Error loading time target for "+category+" "+id)
                label.set_markup("<b>"+str(category)+" Error</b>: '"+str(id)+"' didn't match anything " )
            else:     
                # print('Show time target',item_data)
                label.set_markup("<b>"+item_data['label'] +"</b> "+str(data['type'])+ " "+str(data['value'])+" minutes per "+str(data['within_value'])+" "+str(data['within_unit']) )

                edit_button = Gtk.Button(label="Edit")
                edit_button.connect("clicked", self.edit_add_target, category, id)
                box.pack_end(edit_button,False, False, 5)    
                
            box.add(label)
            
            del_button = Gtk.Button(label="Remove")
            del_button.connect("clicked", self.delete_time_target, category, id)
            box.pack_end(del_button,False, False, 5)

            inner_box.pack_start(box,True,True,5)

            separator = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
            inner_box.pack_start(separator, True, True, 0)

        outer_box.show_all()
  

    def show_connectors(self, connector_category = 'todolists'):
        print("show_connectors: "+connector_category)

        outer_box = self.boxes[connector_category+"_outer"]

        self.boxes[connector_category+"_inner"].destroy()

        self.boxes[connector_category+"_inner"] = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        inner_box = self.boxes[connector_category+"_inner"]
        outer_box.add(inner_box)

        for id, data in conf.user[connector_category].items():

            box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

            checkbutton = Gtk.CheckButton(label=data['label'])
            checkbutton.set_active(data['status'])
            checkbutton.connect("toggled", update_connector_setting, self.app, connector_category, id, 'status')

            box.pack_start(checkbutton,False,False,5)

            edit_button = Gtk.Button(label="Edit")
            edit_button.connect("clicked", self.edit_add_connector, connector_category, id)
            box.pack_end(edit_button,False, False, 5)        
            
            del_button = Gtk.Button(label="Remove")
            del_button.connect("clicked", self.delete_connector, connector_category, id)
            box.pack_end(del_button,False, False, 5)

            inner_box.pack_start(box,True,True,5)

            separator = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
            inner_box.pack_start(separator, True, True, 0)

        outer_box.show_all()


    def delete_connector(self, widget, connector_category, id):
        connector = conf.user[connector_category][id]
        print("delete "+connector_category+" connector "+ id)
        print(connector)

        if connector_category == 'todolists':
            db_query("DELETE FROM tasks WHERE todolist = ? AND id NOT IN (SELECT task_id FROM sessions)",(id,))
            db_query("DELETE FROM lists WHERE todolist = ?",(id,))

        del conf.user[connector_category][id]
        save_user_settings()
        self.show_connectors(connector_category)


    def delete_time_target(self, widget, category, id):
        target = conf.user['time_targets'][category][id]
        print("delete target ",target)
        del conf.user['time_targets'][category][id]
        save_user_settings()
        self.show_time_targets(category)


    def edit_add_connector(self, widget, connector_category, id = None):

        # If id is set it will edit that connector entry, otherwise add a new one of type in self.inputs['new_'+connector_category+'_type'] 
        if id:
            connector_type = conf.user[connector_category][id]['type']
        else:
            connector_type = self.inputs['new_'+connector_category+'_type'].get_active_text()

        print("type in edit_add_connector %s" % (connector_type))

        dialog = EditAddConnectorDialog(self, connector_category, connector_type, id)
        response = dialog.run()

        if response == Gtk.ResponseType.OK:

            print("The OK button was clicked")
            self.show_connectors(connector_category)

            # TODO: Show connector errors

            self.app.async_refresh()


        elif response == Gtk.ResponseType.CANCEL:
            print("The Cancel button was clicked")

        dialog.destroy()


    def edit_add_target(self, widget, category, id = None):

        # If id is set it will edit that entry, otherwise add a new one of category in self.inputs['new_'+_category+'_type'] 
        # if id:
        #     connector_type = conf.user[connector_category][id]['type']
        # else:
        #     connector_type = self.inputs['new_'+connector_category+'_type'].get_active_text()

        print("category in edit_add_target %s" % (category))

        dialog = EditAddTargetDialog(self, category, id)
        response = dialog.run()

        if response == Gtk.ResponseType.OK:

            print("The OK button was clicked")
            self.show_time_targets(category)


        elif response == Gtk.ResponseType.CANCEL:
            print("The Cancel button was clicked")

        dialog.destroy()


    def on_close(self, obj):
        print("good bie settings window ")
        del self.app.settings_window  










class EditAddTargetDialog(Gtk.Dialog):

    def __init__(self, parent, category, id = None):

        super().__init__(title="Set Time Target", transient_for=parent, flags=0)

        self.set_default_size(600,400)


        if id: 
            data = conf.user['time_targets'][category][id]
            self.item_id = id
            self.initial_item_id = id
        else:
            data = {
                    "type": 'max',
                    "value":15,
                    # "num_days":1,

                    "within_value":1,
                    "within_unit":"days"
            }
            self.initial_item_id = None
            self.item_id = None


        self.entries = {}

        self.set_border_width(15)

        box = self.get_content_area()
        box.set_spacing(15)

        print('target category: '+category)

        box.add(Gtk.Label(label=category))


        if id:
            item_data = db_get_item_by_id(id,category)
            self.menu_bar_item = Gtk.MenuItem(label=item_data['label'])
        else:
            item_data = None
            self.menu_bar_item = Gtk.MenuItem(label="Select "+ category)


        if category == 'tasks':
            menubar = Gtk.MenuBar()
            menubar.append(self.menu_bar_item)
            change_task_menu = Gtk.Menu()
            self.menu_bar_item.set_submenu(change_task_menu)
            add_todos_to_menu(change_task_menu, {}, {}, self.set_item_id,True)
            box.add(menubar)


        elif category == 'lists':
            set_list_widget = choose_from_lists(self.set_item_id, selected_list_id=item_data,accepts_tasks=False )
            box.add(set_list_widget)
        

        self.type_select = Gtk.ComboBoxText()
        self.type_select.append("max","max")
        self.type_select.append("min",'min')
        self.type_select.set_active_id(data['type'])
        box.add(self.type_select)
        
        value_adjustment = Gtk.Adjustment(
            value=int(data['value']),
            lower=0.01,
            upper=9999,
            step_increment=1,
            page_increment=5,
            page_size=0)

        box.add(Gtk.Label(label="Minutes"))
        self.value = Gtk.SpinButton(adjustment=value_adjustment,climb_rate=1,digits=1)

        box.add(self.value)

        box.add(Gtk.Label(label="within this many"))  

        # backward compatibility 
        if 'within_value' not in data:
            data['within_value'] = data['num_days']

                
        within_value_adjustment = Gtk.Adjustment(
            value=int(data['within_value']),
            lower=0.01,
            upper=9999,
            step_increment=1,
            page_increment=5,
            page_size=0)

        self.within_value = Gtk.SpinButton(adjustment=within_value_adjustment,climb_rate=1,digits=1)

        box.add(self.within_value)

        self.within_unit_select = Gtk.ComboBoxText()
        self.within_unit_select.append("days","days")
        self.within_unit_select.append("weeks",'weeks')
        self.within_unit_select.append("months",'months')
        self.within_unit_select.append("years",'years')
        if 'within_unit' in data:
            self.within_unit_select.set_active_id(data['within_unit'])
        box.add(self.within_unit_select)

        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )

        self.show_all()
        self.connect("response", self.on_response, category)


    def set_item_id(self,w=None,data=None):
        print("set_task",data)
        self.item_id = data['id']
        self.menu_bar_item.set_label(data['label'])


    def on_response(self, widget, response, category):

        if response == Gtk.ResponseType.OK:

            if self.initial_item_id and self.initial_item_id != self.item_id:
                del conf.user['time_targets'][category][self.initial_item_id]

            conf.user['time_targets'][category][self.item_id] = {
                'type':self.type_select.get_active_text(),
                'value': self.value.get_value(),
                'within_value': self.within_value.get_value(),
                'within_unit': self.within_unit_select.get_active_text(),
                }
            
            print("save time target",conf.user['time_targets'][category][self.item_id])

            save_user_settings()

        elif response == Gtk.ResponseType.CANCEL:
            print("Cancel button clicked")
            self.entries = {}
        else:
            print("Dialog closed")


            



















class EditAddConnectorDialog(Gtk.Dialog):

    def __init__(self, parent, connector_category, connector_type, id = None):

        super().__init__(title="Connect "+connector_category, transient_for=parent, flags=0)

        self.entries = {}

        self.set_border_width(15)

        box = self.get_content_area()
        box.set_spacing(15)

        print('EditAddConnectorDialog: arguments')
        print('connector_type: ',connector_type)
        print('id: ',id)
        fields = copy.copy(conf.connectors[connector_category][connector_type])
        print('fields',fields)
        
        if(id):
            data = conf.user[connector_category][id]
        else:
            data = fields

        print("connector data",data)

        for key in fields:

            if key in ['id','type','timetracker','status']:
                continue

            label = Gtk.Label(label=key)
            box.add(label)


            if key in ['file']:
                self.entries[key] = Gtk.FileChooserButton(title="Select Todo File")
                self.entries[key].set_action(Gtk.FileChooserAction.OPEN)
                
            else:
                self.entries[key] = Gtk.Entry()
            
            if key in data:

                if key in ['file']:
                    self.entries[key].set_filename(str(data[key]))
                else:
                    self.entries[key].set_text(str(data[key]))

                if key in ['password','token','key']:
                    self.entries[key].set_visibility(False)

            box.add(self.entries[key])

        
        if connector_category == 'todolists':

            box.add(Gtk.Label(label="Time Tracker"))  
            self.timetracker_select = Gtk.ComboBoxText()

            for item in conf.user['timetrackers']:
                self.timetracker_select.append(item,item)

            if 'timetracker' not in data:
                data['timetracker'] = list(conf.user['timetrackers'].keys())[0]
            
            self.timetracker_select.set_active_id(data['timetracker'])
            box.add(self.timetracker_select)


        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )

        self.show_all()
        self.connect("response", self.on_response, connector_category, connector_type, id)


    def on_response(self, widget, response, connector_category, connector_type, id):
        if response == Gtk.ResponseType.OK:

            new = {'type':connector_type,'status':True}

            for key in self.entries:

                if key in ['file']:
                    new[key] = self.entries[key].get_filename()
                else:
                    new[key] = self.entries[key].get_text()
                dbg("value for "+ key +" is "+ new[key] )

            if not id:
                if new['label'] and new['label'] not in conf.user[connector_category]:
                    id = new['label']
                else:
                    # add a 2 char hash
                    import uuid
                    id = new['label']+"-"+str(uuid.uuid4())[:2]
            
            new['id'] = id

            if connector_category == 'todolists':
                new['timetracker'] = self.timetracker_select.get_active_text()

            conf.user[connector_category][id] = new

            save_user_settings()

            # load the connector module 
            conf.todo_connectors[new['type']] = importlib.import_module('connectors.'+new['type'])


        elif response == Gtk.ResponseType.CANCEL:
            print("Cancel button clicked")
            self.entries = {}
        else:
            print("Dialog closed")


            


# pip install caldav
# See https://github.com/python-caldav/
# And https://github.com/python-caldav/caldav/blob/master/examples/basic_usage_examples.py
import caldav
import concurrent.futures
import conf
import utils
import datetime
import uuid

def add_new_task(user_conf,list,task_label):
    uidNew = str(uuid.uuid1())

    with caldav.DAVClient(url=user_conf['url'], username=user_conf['username'], password=user_conf['password']) as client:

        calendar = client.calendar(url=list['id'])
        result = calendar.add_todo(
            summary=task_label,
            # due="Some Date",
            # categories="Some Category",
            id=uidNew
        )
        print('add_task result',result)

        t = {
            'id':uidNew,
            'label':task_label,
            'parent_id':list['id'],
            'parent_label':list['label'],
            'todolist':user_conf['id'],
            'data':{}
        }

        return t
        
def mark_task_done(task):

    print("Mark "+task['label']+" done" )
    user_conf = conf.user['todolists'][task['todolist']]

    with caldav.DAVClient(url=user_conf['url'], username=user_conf['username'], password=user_conf['password']) as client:

        calendar = client.calendar(url=task['parent_id'])
        print("got calendar")
        print(calendar)

        taskFetched = calendar.search(
            todo=True,
            uid=taskToModify["UID"],
            )


        # Extremely slow, not sure if it's the client or the server implementation 
        for t in calendar.todos():
            if str(t.icalendar_component["uid"]) == task['id']:
                print("found the task!")
                print(t.icalendar_component)
                print("t in mark done calendar")
                t.complete()
                return True

        # How it's supposed to work but doesn't 
        # # event = caldav.Event(client=client, parent=calendar, url=task['id'])
        # todo = caldav.Todo(client=client, parent=calendar, url=task['id'])

        # print("got 'todo' ?")
        # print(todo)
        # todo.load()
        # print('\nSummery after load()')
        # print(todo.icalendar_component["summary"])

        # result = todo.complete()

# def load_calendar_todo_data(i):
    
def load_calendar_todo_index(i):
    # print('load_calendar_todos',i)
    
    c = i['c']
    user_conf = i['user_conf']

    
    start_time = utils.now()

    print(c.name, ': start caldav todos in thread')
    # caldav_tasks = c.todos()

    caldav_tasks = c.search(
    #     # start=datetime.today(),
    #     # end=datetime(date.today().year + 5, 1, 1),
    #     # event=True,
        todo=True,
        expand=False,
    )
    todo_function_time = utils.now()

    print(c.name, ': caldav calendar.search took',utils.now() - start_time )

    tasks = {}
    for t in caldav_tasks:
        tasks[str(t.icalendar_component["uid"])] = {
                'id':str(t.icalendar_component["uid"]),
                'label':str(t.icalendar_component["summary"]),
                'parent_label':str(c.name),
                'parent_id':str(c.url),
                'todolist':user_conf['id'],
                'data':{}
            }
        # print('speed test ')
        # print(t.data)t
        # print(t.icalendar_component)
        # print("loaded: "+ str(t.icalendar_component["summary"]))
    total_time = utils.now() - start_time
    print(c.name, ':',len(caldav_tasks), ' tasks, total time ',total_time )
    return tasks


def get_todos(user_conf):

    tasks = {}
    lists = {
        user_conf['id']: {
            'id':user_conf['id'],
            'label':user_conf['label'],
            'parent_id':'',
            'parent_label':'',
            'todolist':user_conf['id'],
            'status': 1,
            'data':{
                'accepts_tasks':False
            }
        }
    }

    with caldav.DAVClient(url=user_conf['url'], username=user_conf['username'], password=user_conf['password']) as client:

        principal = client.principal()
        calendars = principal.calendars()
            
        print("loading %i CalDav lists:" % len(calendars))
        params = []
        
        # For faster testing
        # calendars = [calendars[1]]
        # calendars = [calendars[1],calendars[2]] 
        for c in calendars:
            print(c.name +" url:"+str(c.url))
        
            lists[c.url] = {
                'id':str(c.url),
                'label':str(c.name),
                'parent_id':user_conf['id'],
                'parent_label':user_conf['label'],
                'todolist':user_conf['id'],
                'status':1,
                'data':{
                    'accepts_tasks':True,
                }
            }

            params.append({'c':c,'user_conf': user_conf})

        # print('params',params)


        with concurrent.futures.ThreadPoolExecutor() as executor:
           result = executor.map(load_calendar_todo_index, params)

        for list_tasks in result:
            tasks.update(list_tasks)

    return {'tasks':tasks,'lists':lists}

    


import os
import time
from datetime import datetime, timezone

import csv

import conf

import utils


# csv time tracking 
def save_session(session,user_conf):
    # date,duration,project,task,start time

    session['start_time'] = session['start_time'].astimezone()

    row = [session['start_time'].date(),str(round(session['duration']) / 60) +" minutes",session['task']['parent_label'],session['label'],session['start_time'].strftime('%H:%M:%S')]
    # row = [session['start_time'].date(),(round(session['duration']) / 60),session['parent'],session['label'],session['start_time'].time()]
    # row = [session['start_time'].date(),(round(session['duration']) / 60),session['label'],session['start_time'].time()]


    with open(user_conf['file'],'a') as f:
        writer = csv.writer(f)
        writer.writerow(row)



import os
import datetime

import pytodotxt

import conf
import utils 

# lists and tasks are each a dict of dicts with following shape 
# tasks[task_id] = {
#     'id':task_id,
#     'label':label,
#     'parent_id':parent_id,
#     'parent_label':parent_label,
#     'status':1, #( 1 pending, 0 done, -1 deleted)
#     'priority':1 max, 5 min, 0 none,
#     'todolist':user_conf['id'],
#     'data':{
#         # Any additional data
#     }
# }

def watch(user_conf):
    import pyinotify

    """Watch for modifications of the todo file with pyinotify."""
    wm = pyinotify.WatchManager()
    notifier = pyinotify.ThreadedNotifier(wm, mark_changed, user_conf)
    notifier.start()

    # wm.add_watch(os.path.dirname(user_conf['file']), pyinotify.IN_MODIFY | pyinotify.IN_MOVED_TO) # Watch whole folder. Perhaps more reliable 
    wm.add_watch(user_conf['file'], pyinotify.IN_MODIFY | pyinotify.IN_MOVED_TO) # just the file. Perhaps better 


def mark_changed(event, user_conf):
    ''' This is in a thread '''
    if event.pathname == user_conf['file']:
        conf.todo_sync_required[user_conf['id']] = True
    

def add_new_task(user_conf,list,task_label):
    ''' Adds the label to the provided list in the specified todo and returns a task dict'''
    todotxt = pytodotxt.TodoTxt(user_conf['file'])
    todotxt.parse()

    task = pytodotxt.Task(task_label +"+"+list['label'])

    todotxt.add(task)
    todotxt.save()
    # task.add_project

    t = {
            'id':task_label,
            'label':task_label,
            'parent_id':list['label'],
            'parent_label':list['label'],
            'status':1,
            # 'priority':1 max, 5 min, 0 none,
            'todolist':user_conf['id'],
            'data':{
            }
        }
    return t
            
def mark_task_done(task):
    '''Return True on success False on error'''
    
    todotxt = pytodotxt.TodoTxt(conf.user['todolists'][task['todolist']]['file'])

    for t in todotxt.parse():
        if t.bare_description() == task['id']:
            t.completion_date = datetime.date.today()
            t.is_completed = True
            break

    todotxt.save()


def get_todos(user_conf):
    '''Return {'lists':lists,'tasks':tasks} '''

    tasks = {}
    lists = {}

    lists[user_conf['id']] = {
            'id':user_conf['id'],
            'label':user_conf['label'],
            'parent_id':'',
            'parent_label':'',
            'todolist':user_conf['id'],
            'status': 1,
            'data':{
                'accepts_tasks':True
            }
        }


    todotxt = pytodotxt.TodoTxt(user_conf['file'])
    for t in todotxt.parse():
        
        tasks[t.bare_description()] = {
            'id':t.bare_description(),
            'label':t.bare_description(),
            'parent_id':user_conf['id'],
            'parent_label':user_conf['label'],
            'status':1,
            # 'priority':1 max, 5 min, 0 none,
            'todolist':user_conf['id'],
            'data':{
            }
        }

        if t.is_completed:
            tasks[t.bare_description()]['status'] = 0


        if t.projects:
            l = t.projects[0]
            tasks[t.bare_description()]['parent_id'] = l
            if l not in lists:
                lists[l] = {
                    'id':l,
                    'label':l,
                    'parent_id':user_conf['id'],
                    'parent_label':user_conf['label'],
                    'status':1, 
                    'priority':0,
                    'todolist':user_conf['id'],
                    'data':{
                        'accepts_tasks':True
                    }
                }


    todos =  {'lists':lists,'tasks':tasks}

    return todos

# testing
# print(get_todos())


import json
import time
from datetime import datetime, timezone
from urllib.request import urlopen
import requests
from dateutil import tz

import conf

from utils import time_to_sec, sec_to_time, now, time_difference, open_external, get_lists_for_item,  lists_cache


def save_session(session,user_conf):

    # TODO: fix this (seems to create "/ FINAL_LIST ")
    tree = " / ".join(get_lists_for_item(session['task'])['labels'])

    payload = {
        "timestamp":session['start_time'].replace(tzinfo=tz.tzlocal()).astimezone(timezone.utc).isoformat(), 
        "duration":session['duration'],
        "data":{
            "running":False,
            "label":tree +" / "+ session['label'],
            "id":str(session['task']['id']),
            "project":session['task']['parent_label'],
        }
    }

    print(payload)

    r = requests.post(user_conf['url']+'api/0/buckets/aw-stopwatch/events', json=payload)



import os.path
import json
import importlib
from datetime import datetime, timezone
from pathlib import Path
import copy

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, Gdk

# Set working dir to file location 
os.chdir(os.path.dirname(os.path.realpath(__file__)))
app_id = "what-am-i-doing" # Folder name for xdg config and data locations 
app_name = "What am I Doing?" # What shows up with pidof and in window titles etc


user_data_dir = GLib.get_user_data_dir()+"/"+app_id
Path(user_data_dir).mkdir(parents=True, exist_ok=True)

user_settings_dir = GLib.get_user_config_dir()+"/"+app_id
Path(user_settings_dir).mkdir(parents=True, exist_ok=True)

# key and type must be the same, (Seems redundant but it's is quite helpful) 
connectors = {
    "todolists":{
        'txt':{'id':'txt','type':'txt','label':'','file':'',"status":True,'timetracker':''},
        'vikunja':{'id':'vikunja','type':'vikunja','label':'Vikunja','url':'http://localhost:3456/', 'token':'',"username":"","status":True,'timetracker':''},
        'caldav':{'id':'caldav','type':'caldav','label':'CalDav Todo','url':'http://localhost:3456/','password': "",'username': '',"status":True,'timetracker':''},
        'psc_timetracker':{'id':'psc_timetracker','type':'psc_timetracker','label':'Timetracker','url':'https://photosynth.ca/timetracker/','key':'',"status":True,'timetracker':'Timetracker'},
        'taskwarrior':{'id':'taskwarrior','type':'taskwarrior','label':'TaskWarrior',"status":True,'timetracker':''},
        'todotxt':{'id':'todotxt','type':'todotxt','label':'','file':"","status":True,'timetracker':''},
    },
    "timetrackers":{
        'csv':{'id':'csv','type':'csv','label':'CSV file','file':'sessions.csv',"status":True}, 
        'activitywatch':{'id':'activitywatch','type':'activitywatch','label':'ActivityWatch','url':'http://localhost:5600/',"status":True}, 
        'psc_timetracker':{'id':'psc_timetracker','type':'psc_timetracker','label':'Timetracker','url':'https://photosynth.ca/timetracker/','key':'',"status":True},
        'timewarrior':{'id':'timewarrior','type':'timewarrior','label':'TimeWarrior',"status":True}, 
    }
}

todo_sync_time = datetime.now() 

todo_sync_required = {}

prototype_settings = {
        "pomodoro_interval": 26,
        "randomness_interrupt_interval":5,
        "default_text": "What am I doing?",
        "todolist_refresh_interval":6,
        "version":0.2,
        "display_todolist_as_top_level_list":'auto',
        'max_top_level_menu_items':18,
        'tick_interval':1,

        "custom_pomodoro_intervals": {
            "email":7
        },

        'prompts':'What am I doing?\nWhy am I here?\nWhat could I do?\nWhat do I wish to accomplish?\nWhat is my aim?\nWhat\'s next',

        'time_targets':{
            'tasks':{},
            'lists':{}
        },

        "todolists": {
            "Example Todo":{
                "type": "txt",
                "label": "Example Todo",
                "file": "example-todo.txt",
                "timetracker":"Example CSV",
                "status":True
            }
        },
        "timetrackers":{
            "Example CSV":{
                "label": "Example CSV",
                "type":"csv",
                "file":"sessions.csv",
                "status":True
            }
        }
    }

if os.path.isfile(user_settings_dir+"/user_settings.json"):
    with open(user_settings_dir+"/user_settings.json", "r") as file:
        settings_file = json.load(file)
else:
    print("Setting up initial settings file")
    settings_file = {}


user = {}

# Merge user_settings.json with prototype settings
for key, val in prototype_settings.items():

    if settings_file and key in connectors:
        user[key] = {}

        # merge and validate todolist and timetracker settings against connector prototypes 
        for c_key, c_val in settings_file[key].items():
            
            if c_val['type'] in connectors[key]:
                proto = copy.copy(connectors[key][c_val['type']])
                user[key][c_key] = proto
            else:
                print('ERROR: no connector for type '+c_val['type']+' in user_settings.json','skipping that connector',c_val)
                continue

            # print('validate user_settings connector against prototype.','user',c_val,'proto',proto)

            for p_field, p_value in proto.items():
                
                if p_field in c_val: 
                    if type(c_val[p_field]) == type(p_value):
                        user[key][c_key][p_field] = c_val[p_field]
                    else:
                        print(p_field+" is the wrong type "+str(type(c_val[p_field]))+"("+str(c_val[p_field])+") in user conf for connector "+c_key+" using fallback: "+str(p_value))
                else:
                    print("Adding missing: connection field "+p_field+")") 

    elif key in settings_file and type(settings_file[key]) == type(val):
        user[key] = settings_file[key]
    else:
        user[key] = val
        
# print(json.dumps(user, indent=4))
print(json.dumps(connectors['todolists'], indent=4))

# update time_target format 
for id, tt in user['time_targets']['lists'].items():
    if 'within_value' not in tt:
        print("Updating time target to new format ",tt)
        tt['within_value'] = tt['num_days']
        tt['within_unit'] = 'days'
        print(tt)

for id, tt in user['time_targets']['tasks'].items():
    tt
    if 'within_value' not in tt:
        print("Updating time target to new format ",tt)
        tt['within_value'] = tt['num_days']
        tt['within_unit'] = 'days'
        print(tt)

with open(user_settings_dir+"/user_settings.json","w") as settings_file:
    json.dump(user, settings_file)


todo_connectors = {}
timetracker_connectors = {}

# Load used connector modules
for id, todolist in user['todolists'].items(): 
    todo_connectors[todolist['type']] = importlib.import_module('connectors.'+todolist['type'])
    
for id, timetracker in user['timetrackers'].items(): 
    timetracker_connectors[timetracker['type']] = importlib.import_module('connectors.'+timetracker['type'])



import os
import time
from datetime import datetime, timezone, timedelta

import conf

import utils

from timew import TimeWarrior

timew = TimeWarrior()

# timew time tracking 
def save_session(session,user_conf):

    global timew

    out = {}

    out['start_time'] = session['start_time'].astimezone()    

    out['tags'] = [session['task']['parent_label'],session['label']]
    
    out['tags'] = list(filter(None, out['tags']))
    
    out['end_time'] = out['start_time'] + timedelta(seconds=session['duration'])    
    
    utils.dbg(out,s="timewarrior",l=2)
    
    timew.track(start_time=out['start_time'], end_time=out['end_time'], tags=out['tags'])
 



import requests
import json
from datetime import datetime, timezone, timedelta
import uuid

import conf
import utils

from utils import time_to_sec, sec_to_time, now, time_difference, open_external, get_lists_for_item, lists_cache


def item_is_from_psc_timetracker_todolist(i):
    if conf.user['todolists'][i['todolist']]['type'] == 'psc_timetracker':
        return True
    else:
        utils.dbg('Item is not from a psc_timetracker todolist',i,s="psc_timetracker",l=0)
        return False


def mark_task_done(task):
    utils.error_notice("Sorry, you can't mark this task as done","This feature doesn't seem to be working with PSC timetracker tasks right now. Sorry! You'll have to open timetracker and use the web interface.")
    user_conf = conf.user['todolists'][task['todolist']]
    
    utils.dbg(user_conf['label']+" " +task['label']+" mark_task_done",s="psc_timetracker")

    data = assemble_task_synch_data(task,{'status':'complete'})
    utils.dbg('NEW system data',data, s="psc_timetracker")

    # lists = lists_cache()
    # data = task['data']
    # list_ids = get_lists_for_item(task)['ids']
    # data['clients'][list_ids[1]]['projects'][list_ids[2]]['tasks'][task['id']]['status'] = 'complete'
    
    # utils.dbg('OLD system data',data, s="psc_timetracker")
    # utils.dbg(data, s="psc_timetracker")

    response = requests.post(user_conf['url']+"synch.php?action=synchToServer&key="+user_conf['key'],json=data)
    
    print(response.json())
    return response


def add_new_task(user_conf,parent_list,task_label):
    
    id = str(uuid.uuid4())

    t = {
        'id':id,
        'label':task_label,
        'parent_id':parent_list['id'],
        'parent_label':parent_list['label'],
        'todolist':user_conf['id'],
        'data':{

        }
    }

    # data = {
    #     'clients':{
    #         parent_list['parent_id']:{
    #             'id':parent_list['parent_id'],
    #             'name':parent_list['parent_label'],
    #             'projects':{
    #                 parent_list['id']:{
    #                     'id':parent_list['id'],
    #                     'name':parent_list['label'],
    #                     'tasks':{
    #                         id:{
    #                         'id':id,
    #                         'name':task_label
    #                         }
    #                     }
    #                 }
    #             }
    #         }
    #     }
    # }
    # utils.dbg("old system synchToServer data", data, s="psc_timetracker",l=1)
   
    data = assemble_task_synch_data(t)
    utils.dbg("new system synchToServer data", data, s="psc_timetracker",l=3)

    response = requests.post(user_conf['url']+"synch.php?action=synchToServer&key="+user_conf['key'],json=data)
    
    print(response.json())

    # TODO: check what success and fail messages are returned and check for those
    if response.status_code != 200:
        error_notice("Failed to add tasks","")
        dbg('psc timetracker add task error',response,l=0,s='psc_timetracker')
        return False

    print("Add task "+task_label+" to "+str(parent_list['id'])+" Response",t)

    return t


def assemble_task_synch_data(t, passed_task_data = None):
    ''' return a dict collected from 'project' level list with client, project, task, {t['id']: passed_task_data)'''

    p = utils.db_get_item_by_id(t['parent_id'],'lists')

    list_ids = get_lists_for_item(t)['ids']

    print("psc assemble_task_synch_data")
    
    o = p['data']['sync_obj']

    task_data = {'id':t['id'],'name':t['label']} 

    if type(passed_task_data) == dict:
        task_data = task_data | passed_task_data

    o['clients'][list_ids[1]]['projects'][list_ids[2]]['tasks'] = {t['id']: task_data } 

    return o


def save_session(session,user_conf):
    print('Save Session to PSC timtracker')
    # print(session)
    # print(user_conf)

    # Not a great test but probably sufficient
    if not item_is_from_psc_timetracker_todolist(session['task']):
        
        error_notice("TimeTracker Error","Can't log sessions to psc_timetracker unless the the task is from a timetracker list")
        # TODO: create new client/project/task object with reproducibly-hashed list names as ids?
        return False
    
    # print(session['task'])
    task = session['task']

    session_id = str(uuid.uuid4())

    end_time = session['start_time'] + timedelta(seconds=session['duration'])

    session = {
        session_id:{
            "id": session_id,
            "start_time": session['start_time'].strftime("%Y-%m-%d %H:%M:%S"),
            "end_time": end_time.strftime("%Y-%m-%d %H:%M:%S"),
            "notes": ""
        }
    }

    # old, Working, data attached to every task, version 
    # data = task['data']
    # list_ids = get_lists_for_item(task)['ids']
    # data['clients'][list_ids[1]]['projects'][list_ids[2]]['tasks'][task['id']]['sessions'] = session
    # print("old version synchToServer data",data)

    data = assemble_task_synch_data(task,{'sessions':session})
    # print("NEW version synchToServer data",data)

    # print('session')
    # print(session)
    # print('modified data')
    # print(data)

    response = requests.post(user_conf['url']+"synch.php?action=synchToServer&key="+user_conf['key'],json=data)
    
    print(response.json())
    if response.status_code == 200:
        return True
    else:
        return False

def get_todos(user_conf):

    tasks = {}
    lists = {
        user_conf['id']: {
            'id':user_conf['id'],
            'label':user_conf['label'],
            'parent_id':'',
            'parent_label':'',
            'todolist':user_conf['id'],
            'status': 1,
            'data':{
                'accepts_tasks':False
            }
        }
    }
    # TODO: detect and raise/hande connection error here 
    data = requests.get(user_conf['url']+"synch.php?action=synchFromServer&key="+user_conf['key']).json()
    
    for client_id in data['clients']:
        client = data['clients'][client_id]

        # This is implemented in the database but not the api
        # if client['status'] == 'completed':
        #     continue

        lists[client['id']] = {
            'id':client['id'],
            'label':client['name'],
            'parent_id':user_conf['id'],
            'parent_label':user_conf['label'],
            'todolist':user_conf['id'],
            'data':{
                'accepts_tasks':False
            }
        }

        print(client['name'])

        for p_id in client['projects']:

            p = client['projects'][p_id]
            # This is implemented in the database but not the api
            # if p['status'] == 'completed':
            #     continue
            
            if 'tasks' in p:

                lists[p['id']] = {
                    'id':p['id'],
                    'label':p['name'],
                    'parent_id':client['id'],
                    'parent_label':client['name'],
                    'todolist':user_conf['id'],
                    'data':{
                        'accepts_tasks':True,
                        'sync_obj':{
                            'clients':{ 
                                client['id']:{
                                    'id':client['id'],
                                    'name':client['name'],
                                    'projects':{
                                        p['id']:{
                                            'id':p['id'],
                                            'name':p['name'],
                                            
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }

                print('    '+p['name'])
            
                for t_id in p['tasks']:
                    t = p['tasks'][t_id] 
                    
                    if t['status'] == 'completed' or t['status'] == 'on hold':
                        # continue 
                        status = 0
                    else:
                        status = 1

                    if 'sessions' in t:
                        del t['sessions']

                    print('        '+t['name'])

                    tasks[t['id']] = {
                        'id':t['id'],
                        'label':t['name'],
                        'parent_id':p['id'],
                        'parent_label':p['name'],
                        'todolist':user_conf['id'],
                        'status':status,
                        # What has to be sent back to update the task or add sessions
                        'data':{
                            'clients':{
                                client['id']:{
                                    'id':client['id'],
                                    'name':client['name'],
                                    'projects':{
                                        p['id']:{
                                            'id':p['id'],
                                            'name':p['name'],
                                            'tasks':{
                                                t['id']:t
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }

                    if t['priority'].isnumeric():
                        tasks[t['id']]['priority'] = float(t['priority'])
                    
    todos = {'lists':lists,'tasks':tasks}
    return todos


import os
import time
import json
import webbrowser
import subprocess, sys
from datetime import datetime, timezone, timedelta
from dateutil.relativedelta import relativedelta
import copy

import sqlite3 
from contextlib import closing
from pathlib import Path

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, Gdk
gi.require_version('Notify', '0.7')
from gi.repository import Notify as notify

import conf
notify.init(conf.app_name)

lists = {}

db_file = conf.user_data_dir+"/data.db"

# debug_level = 0 # production value 
debug_level = 2 # dev value 
debug_systems = []



def dbg(*data, **kwargs):
    ''' Any number of positional args then kwargs 
    l=level {0:Error, 1:Warning, 2:Info, 3:Details}  
    s=system (Debuggable systems: 'taskwindow','signals','todoloading','user_settings', 'targets')'''

    levels = {0:'Error', 1:'Warning', 2:'Info', 3:'Details'}

    # default system and level
    system = ''
    level = 2

    if 's' in kwargs: system = kwargs['s']

    if 'l' in kwargs: level = int(kwargs['l'])

    if "all" in debug_systems or system in debug_systems or level <= debug_level: 

        if 'notify' in kwargs:
            # Will this cause a segfault if called from a bg thread?
            notify.Notification.new(conf.app_name+" "+levels[level], str(kwargs['notify']), None).show()

        # o = "Debug "+system+" "+levels[level], tuple(filter(None,data))
        o = system+" "+levels[level], tuple(filter(None,data))
        pretty_print(o)

    # else:
    #     # debug the debug
    #     print("Not displaying dbg ")
    #     # pretty_print(data)


def error_notice(title, details = None, e = None):
    print('ERROR',title,details,e)
    notify.Notification.new(str(title),str(details), None).show()


def pretty_dict(i, item_sep = ", ", use_indents=True, indent=''):
    o = ''
    if use_indents:
        o += indent

    if isinstance(i,dict):
        for key, val in i.items():

            o += str(key).capitalize() + ": "
            if isinstance(val,(dict,list,tuple)): 
                o += "\n" + pretty_dict(val, item_sep, use_indents, indent + '    ')
            else: 
                o += str(val) + item_sep

    elif isinstance(i,(list,tuple)):
        for val in i:
            if isinstance(val,(dict,list,tuple)): 
                o += "\n" + pretty_dict(val, item_sep, use_indents, indent + '    ')
            else: 
                o += str(val) + item_sep
    else:
        o += str(i) + item_sep

    return o


def pretty_print(i):
    print(pretty_dict(i))


def time_to_sec(time=None):
    if not time:
        return 0
    sec = 0
    for k, v in enumerate(reversed(time.split(':'))):
        sec += int(v) * 60 ** k
    return sec

def sec_to_time(sec):

    try:
        int(sec)
    except Exception as e:
        print("sec to time error", e)
        sec = 0

    time = str("{:02d}".format(int(sec // 3600))) + ':' + str("{:02d}".format(int((sec % 3600) // 60))) + ':' + str("{:02d}".format(int(sec % 60)))
    return time.lstrip('0').lstrip(':')

def now():
    return datetime.now()
    # return datetime.now(timezone.utc)
    # return datetime.now(timezone.utc).astimezone().tzinfo

def start_of_day():
    return now().replace(hour=0,minute=0,second=0,microsecond=0)

def time_difference(earlier, later = None):
    ''' Return seconds (float), later defaults to now'''
    # date_format = "%Y-%m-%dT%H:%M:%S.%f%z" # This can probably be updated / tossed (ActivityWatch format )
    if type(earlier) == str:
        earlier = datetime.strptime(earlier,'%Y-%m-%dT %H:%M:%S')

    if not later:
        later = now() 

    difference = later - earlier

    return difference.total_seconds()


def get_connector_openable(widget = None, connector_dict = None, open_it = True):

    for option in ['file','front_end_url','url','uri','open_command']:

        if option in connector_dict:
            if open_it:
                open_external(connector_dict[option])
                return True
            else:
                return connector_dict[option]

    print('ERROR: get_connector_openable failed to find an openable')
    print(connector_dict)
    return False

def open_external(i):
    # if platform.system() == "Windows":
    #     os.startfile(folder_path)
    # else:
    if(i.startswith('http')):
        webbrowser.open(i,1,True)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, i])



def get_lists_for_item(i): 
    ''' i can be a task or a list
    Returns {'ids':['l_1','l_12','l_123],'labels':['todolist label','project','sub project']}'''

    lists = lists_cache()
    # pretty_print(lists)

    ids = []
    labels = []

    if i and 'parent_id' in i and i['parent_id'] and i['parent_id'] in lists:

        l = lists[i['parent_id']]

        while l:
            ids.insert(0, l['id'])
            labels.insert(0, l['label'])
            if l['parent_id'] in lists:
                l = lists[l['parent_id']]
            else: 
                break

    # This may need empty val filtering . see: https://stackoverflow.com/a/3845449/4692205

    output = {'ids':ids,'labels':labels}

    # print('get_lists_for_item',i['label'])
    # pretty_print(output)

    return output


def extended_label(i):
    if not i:
        dbg('emptiness passed to extended_label',i)
        return ''

    lists = get_lists_for_item(i)
    l = lists['labels']
    l_ids = lists['ids']

    if conf.user['display_todolist_as_top_level_list'] != 'always':
        if l_ids and l_ids[0] == i['todolist']:
            del l[0]

    l.append(i['label'])
    o = ' > '.join(l)
    return o


def lists_cache(new_lists = None):
    ''' This is useful but very gimpy. If new_lists is supplied it will exclusively use that (until it is replaced or the global lists var is emptied)'''
    global lists
    
    if(new_lists):
        lists = new_lists
        # TODO: add this to the existing rather than overwriting...
    
    elif(lists == {}):
        lists = db_query('SELECT * FROM lists WHERE status = 1 ORDER BY extended_label DESC',None, 'id') # TODO: limit to active todolists

        print('list_cache l',lists)
        for l in lists.items():
            lists[l['id']]['data'] = json.loads(l['data'])

    return lists 


def first(i, default = None):

    if isinstance(i, (tuple,list)):
        return i[0]

    elif isinstance(i,dict):
        return  next(iter(i.values()))


def save_user_settings():
    print("Save updated user_settings")
    dbg(conf.user)
    with open(conf.user_settings_dir+"/user_settings.json","w") as settings_file:
        json.dump(conf.user, settings_file)


def db_query(sql,parameters=None,key=None,error_handling=1):
    '''error_handling: 2 = raise error, 1 = notify error, 0 = ignore errors '''
    # print('db_query parameters type: ',type(parameters))
    # print('first(parameters)',first(parameters))
    global db_file   
    c = sqlite3.connect(db_file)
    c.row_factory = sqlite3.Row
    try:
        with c:
            if parameters:
                
                if isinstance(parameters, (dict,list)) and isinstance(first(parameters), (dict)):
                    if isinstance(parameters, dict):
                        parameters = list(parameters.values())
                    
                    rows = c.executemany(sql,parameters).fetchall()
                        
                else: 
                    rows = c.execute(sql,parameters).fetchall()

            else:
                rows = c.execute(sql).fetchall()

            result = [dict(row) for row in rows]

            if key and result and key in result[0]:
                o = {}
                for row in result:
                    o[row[key]] = row
                result = o
            # print("db_query result",result)
        c.close()
        return result

    except Exception as e:
        dbg(e,"sql",sql,'parameters',parameters,s="db",l=0)
        if error_handling > 0:
            error_notice("database error",str(e))
            if error_handling > 1:
                raise e
        if key: 
            return {}
        else:
            return []


def db_init():
    if not os.path.isfile(db_file):
        print('initializing database')

        db_query("CREATE TABLE lists (id TEXT, label TEXT DEFAULT '', parent_id TEXT DEFAULT '', parent_label TEXT DEFAULT '', todolist TEXT DEFAULT '', priority INTEGER DEFAULT 0, status INTEGER DEFAULT 1, extended_label TEXT DEFAULT '', data TEXT DEFAULT '{}');")

        db_query("CREATE TABLE tasks (id TEXT, label TEXT DEFAULT '', parent_id TEXT DEFAULT '', parent_label TEXT  DEFAULT '', todolist TEXT DEFAULT '', priority INTEGER DEFAULT 0, status INTEGER DEFAULT 1, extended_label TEXT, data TEXT DEFAULT '{}');")
        
        db_query("CREATE TABLE sessions (start_time TEXT, duration INTEGER, task_id TEXT, parent_id TEXT, todolist TEXT, extended_label TEXT, notes TEXT, timetracker TEXT);")

        db_query("CREATE TABLE system (field TEXT PRIMARY KEY NOT NULL, value TEXT);")

        db_query("INSERT INTO system(field, value) VALUES('db_schema_version', '0.4')")
        

def db_update():

    try:
        db_schema_version = float(db_query("SELECT field, value FROM system WHERE field = 'db_schema_version' ",error_handling=2)[0]['value'])
    except Exception as e:
        print("Updating db_schema_version to 0.2")

        db_query("CREATE TABLE system (field TEXT PRIMARY KEY NOT NULL, value TEXT)")
        db_query("REPLACE INTO system(field, value) VALUES('db_schema_version', '0.2')")

        db_schema_version = float(db_query("SELECT field, value FROM system WHERE field = 'db_schema_version' ")[0]['value'])

        db_query("ALTER TABLE lists ADD COLUMN status INTEGER DEFAULT 1")
        db_query("ALTER TABLE tasks ADD COLUMN status INTEGER DEFAULT 1")

        db_query("ALTER TABLE lists ADD COLUMN extended_label TEXT")
        db_query("ALTER TABLE tasks ADD COLUMN extended_label TEXT")

        db_query("ALTER TABLE sessions ADD COLUMN extended_label TEXT")

        # add default values for priority and data 
        
        # Since these column are just a cache, replace instead of copying columns
        db_query("ALTER TABLE tasks DROP COLUMN priority")
        db_query("ALTER TABLE tasks DROP COLUMN data")
        db_query("ALTER TABLE lists DROP COLUMN priority")
        db_query("ALTER TABLE lists DROP COLUMN data")

        # add new columns
        db_query("ALTER TABLE lists ADD COLUMN priority INTEGER DEFAULT 0")
        db_query("ALTER TABLE lists ADD COLUMN data TEXT DEFAULT '{}'")
        db_query("ALTER TABLE tasks ADD COLUMN priority INTEGER DEFAULT 0")
        db_query("ALTER TABLE tasks ADD COLUMN data TEXT DEFAULT '{}'")


    if db_schema_version == 0.2:
                
        db_query("ALTER TABLE lists DROP COLUMN status")
        db_query("ALTER TABLE tasks DROP COLUMN status")

        db_query("ALTER TABLE lists ADD COLUMN status INTEGER DEFAULT 1")
        db_query("ALTER TABLE tasks ADD COLUMN status INTEGER DEFAULT 1")

        db_query("REPLACE INTO system(field, value) VALUES('db_schema_version', '0.3')")
    
        db_schema_version = 0.3


    if db_schema_version == 0.3:
        db_query("ALTER TABLE sessions ADD COLUMN timetracker TEXT")
        db_query("ALTER TABLE sessions ADD COLUMN notes TEXT")
        db_query("REPLACE INTO system(field, value) VALUES('db_schema_version', '0.4')")
        print('adding timetracker column to session table')

        # for s in db_query("SELECT start_time, task_id, todolist FROM sessions"):

        for todolist_id, todo in conf.user['todolists'].items():
            print("set timetracker to ",todo['timetracker'], " for sessions from ",todolist_id)

            db_query("UPDATE sessions SET timetracker = ? WHERE todolist = ?",(todo['timetracker'],todolist_id) )

        db_schema_version = 0.4

    # if db_schema_version == 0.4:


    dbg('db_schema_version', db_schema_version,s='db')

def db_to_py_type_map(db_type, value=None):
    types = {
        'INTEGER':int,
        'BOOLEAN':bool,
        'REAL':float,
        'TEXT':str,
        'BLOB':bytes,
    }

    if isinstance(value, (dict,list,tuple)) and db_type in ['TEXT','BLOB']:
        return json.dumps(value)

    if db_type in types: 
        if value:
            return types[db_type](value)
        else:
            types[db_type]


def db_columns(table, cache = {}):
    ''' Returns the PRAGMA query, from cache '''

    if table not in cache: 

        if table not in ['lists','tasks','sessions','system']:
            raise ValueError("bad table passed to db_columns!")
        else:
            fields = db_query("PRAGMA table_info("+table+")")

            # sadly "Parameter markers can be used only for expressions, i.e., values. You cannot use them for identifiers like table and column names." 
            # fields = db_query("PRAGMA table_info(?)",(table,)) 
            
            cache[table] = fields

    return cache[table]


def time_target_priority(i, table = "tasks"):

    if 'priority' in i and i['priority'] > 0:
        # pass through existing priority
        # NOTE: Maybe de-prioritize here if it has surpassed a max target?
        return i['priority']
    else:
        time_target_ballance = check_time_target(i, table, 'min')
        if time_target_ballance and time_target_ballance > 0:
            # This could have more ... nuance...
            i['priority'] = 1 

        else: 
            i['priority'] = 0

    return i['priority'] 


def db_prepare_item(i,table = "tasks"):
    if table not in ['lists','tasks']:
        raise ValueError("bad table passed to db_prepare!")

    fields = db_columns(table)

    o = {}

    if not 'status' in i:
        i['status'] = 1

    if not 'extended_label' in i:
        i['extended_label'] = extended_label(i)

    # time_target_ballance = check_time_target(i, table, 'min')
    # if time_target_ballance and time_target_ballance > 0:
    #     dbg("prioritize min time_target task in db_prepare")
    #     i['priority'] = 1

    # print('fields',fields)
    for f in fields:
        if f['name'] in i:
            o[f['name']] = db_to_py_type_map(f['type'],i[f['name']])
        else:
            o[f['name']] = f['dflt_value']
    
    return o


def db_set_item(i,table = "tasks"):
    i = db_prepare_item(i,table)

    columns = ', '.join(i.keys())
    placeholders = ':'+', :'.join(i.keys())

    db_query('INSERT INTO '+table+' (%s) VALUES (%s)' % (columns, placeholders),[i])


def db_set_todolist(todolist_id,lists,tasks):
    # print('db_set_todolist params: \ntodolist_id: ',todolist_id,'lists: ',lists,'tasks: ',tasks)

    insert_tasks = {}
    insert_lists = {}

    # use the parameter-provided lists for extended label instead of the existing (db) lists
    lists_cache(lists)
        
    for id, i in lists.items():
        insert_lists[id] = db_prepare_item(i,'lists')
    
    for id, i in tasks.items():
        insert_tasks[id] = db_prepare_item(i,'tasks')

    # Clear incomplete lists cache
    lists_cache({})

    # for testing
    # insert_tasks = {'3':insert_tasks['3'],'294':insert_tasks['294']}
    # print('insert_tasks: ',json.dumps(insert_tasks,indent=4))

    db_query("DELETE FROM tasks WHERE todolist = ?",(todolist_id,)) # What about removed 
    db_query("DELETE FROM lists WHERE todolist = ?",(todolist_id,))

    if insert_tasks:
        task_columns = insert_tasks[list(insert_tasks.keys())[0]].keys()
        columns = ', '.join(task_columns)
        placeholders = ':'+', :'.join(task_columns)

        db_query('INSERT INTO tasks (%s) VALUES (%s)' % (columns, placeholders),insert_tasks)
        
    if insert_lists:
        list_columns = insert_lists[list(insert_lists.keys())[0]].keys()
        columns = ', '.join(list_columns)
        placeholders = ':'+', :'.join(list_columns)

        db_query('INSERT INTO lists (%s) VALUES (%s)' % (columns, placeholders),insert_lists)


def db_cleanup():
    
    # ids = tuple(conf.user['todolists'].keys())
    ids = []
    # print('ids_sql',ids_sql)
    for id, todo in conf.user['todolists'].items():
        if todo['status']:
            ids.append(id)

    # Old save
    print("Active todolist ids",ids)

    # Delete unlinked tasks
    db_query('DELETE FROM tasks WHERE todolist NOT IN (%s)' % ','.join('?'*len(ids)),ids)
    # print(len(orphaned_tasks),"orphaned_tasks",orphaned_tasks)
    # orphaned_tasks = db_query('SELECT id, extended_label, todolist FROM tasks WHERE todolist NOT IN (%s)' % ','.join('?'*len(ids)),ids)
    # print(len(orphaned_tasks),"orphaned_tasks",orphaned_tasks)
    
    # Delete unlinked lists
    db_query('DELETE FROM lists WHERE todolist NOT IN (%s)' % ','.join('?'*len(ids)),ids)

    # orphaned_lists = db_query('SELECT id, extended_label, todolist FROM lists WHERE todolist NOT IN (%s)' % ','.join('?'*len(ids)),ids)
    # print("orphaned_lists",orphaned_lists)

    missing_task_sessions = db_query("SELECT * FROM sessions WHERE task_id NOT IN (SELECT id FROM tasks)")
    # print("missing_task_sessions",missing_task_sessions)
    print(len(missing_task_sessions))

    for s in missing_task_sessions:
        # print("missing_task_session")
        # print(pretty_dict(s))
        if s['extended_label']:
            possible_matches = db_query("SELECT * FROM tasks WHERE extended_label = ? ",(s['extended_label'],))
            if possible_matches:
                print("possible_matches")
                print(pretty_dict(possible_matches))

    # update extended_labels
    mismatched_sessions = db_query("SELECT * FROM sessions WHERE extended_label NOT IN (SELECT extended_label FROM tasks)")

    # db_query("UPDATE sessions SET extended_label = (SELECT extended_label FROM tasks WHERE id = task_id AND parent_id = parent_id) WHERE task_id IN (SELECT id FROM tasks) AND extended_label NOT IN (SELECT extended_label FROM tasks)")



    # SELECT * FROM sessions WHERE extended_label IS NULL;
    # UPDATE sessions SET extended_label = (SELECT extended_label FROM tasks WHERE id = task_id) WHERE extended_label IS NULL

    # db_query("SELECT label, todolist FROM tasks WHERE todolist NOT IN ?",(ids_sql,))
    # db_query("DELETE FROM tasks WHERE todolist  NOT IN "+ids_sql)
    # db_query("DELETE FROM lists WHERE todolist NOT IN "+ids_sql)

    # exit()

def db_get_item_by_id(id,table = 'tasks'):
    
    if table not in ['tasks','lists']:
        dbg('bad table "'+str(table)+'" passed to db_get_item_by_id',id,l=0)
        return {}

    try:
        data = db_query("SELECT * FROM "+table+" WHERE id = ?",(id,))
        # print(data)
        if data:
            return proc_db_item(data[0],table)
        else:
            dbg('db_get_item_by_id from '+table+' failed for id +',id,l=0)
            return {}

    except Exception as e:
        dbg('db_get_item_by_id failed',e,l=0)


def proc_db_item(i,table='tasks'):
    ''' Parse task or list data json and apply time target priority '''
    i['data'] = json.loads(i['data'])
    i['parent_label'] = str(i['parent_label']) #
    i['priority'] = time_target_priority(i,table)
    return i


def db_get_todolist(todolist_id):
    tasks = {}

    for t in db_query("SELECT * FROM tasks WHERE status = 1 AND todolist = ?",(todolist_id,)):
        tasks[t['id']] = proc_db_item(t)

    lists = {}
    for l in db_query("SELECT * FROM lists WHERE todolist = ? AND status = 1",(todolist_id,)):
        lists[l['id']] = proc_db_item(l,table='lists')

    o = {'lists': lists, 'tasks':tasks}
    return o


def db_save_session(session):
    prepared_session = {
        'start_time' : str(session['start_time'].strftime("%Y-%m-%d %H:%M:%S")),
        'duration': round(session['duration']),
        'task_id': str(session['task']['id']),
        'parent_id': str(session['task']['parent_id']),
        'todolist': str(session['task']['todolist']),
        'extended_label': str(session['extended_label']),
        'timetracker': str(session['timetracker']),
        'notes': str(session['notes']),
    } 
    db_query("INSERT INTO sessions(start_time, duration, task_id, parent_id, todolist, extended_label,timetracker, notes) VALUES(:start_time, :duration, :task_id, :parent_id, :todolist, :extended_label, :timetracker, :notes )",prepared_session)


def get_total_time(id, category = 'tasks', start_time = None, end_time = None, get_minutes = None):
    '''If start_time is int or float: it is treated as a number of days,
    Default returns num seconds, (unless get_minutes is True)
    end_time only work if start time is specified '''

    if category in ['list','lists']:
        where = 'parent_id'
    else:
        where = 'task_id'

    if start_time:
        
        if type(start_time) in [int,float]:
            start_time = start_of_day() - timedelta(days = start_time)
         
        if not end_time: end_time = now()
        
        result = db_query("SELECT sum(duration) as duration FROM sessions WHERE "+where+" = ? AND start_time > ? AND start_time < ? ",(id, start_time.strftime("%Y-%m-%d %H:%M:%S"), end_time.strftime("%Y-%m-%d %H:%M:%S")))
    else:
        result = db_query("SELECT sum(duration) as duration FROM sessions WHERE "+where+" = ? ",(id,))

    # TODO make the list version (reverse) recursive
    # print("get_total_time result",result)

    result = force_number(result[0]['duration'])

    if get_minutes:
        result = result and result / 60 or 0
    
    return result


def get_recent_tasks(count = 15):
    data = db_query("SELECT DISTINCT tasks.* FROM tasks JOIN sessions ON sessions.task_id = tasks.id WHERE tasks.status = 1 GROUP BY tasks.id ORDER BY MAX(sessions.start_time) DESC LIMIT ?",(count,))
    o = {}
    for t in data:
        o[t['id']] = proc_db_item(t)
    return o


def get_priority_tasks(count = 100):

    p_tasks = db_query("SELECT DISTINCT tasks.* FROM tasks WHERE priority > 0 AND status = 1 ORDER BY priority ASC LIMIT ?",(count,))

    tt_task_ids = list(conf.user['time_targets']['tasks'].keys())
    tt_list_ids = list(conf.user['time_targets']['lists'].keys())

    # TODO: [Performance] combine these two queries if possible
    tt_tasks = db_query('SELECT DISTINCT * FROM tasks WHERE id IN(%s)' % ','.join('?'*len(tt_task_ids)),tt_task_ids)
    
    tt_list_tasks = db_query('SELECT DISTINCT * FROM tasks WHERE parent_id IN(%s)' % ','.join('?'*len(tt_list_ids)),tt_list_ids)

    o = {}

    for t in p_tasks + tt_tasks + tt_list_tasks: 
    # for t in tt_tasks + tt_list_tasks: #for testing
        i = proc_db_item(t)
        if i['priority'] > 0: 
            o[t['id']] = t
        # else: 
            # dbg('de-prioritize '+ t['label'], i['priority'], s='targets')


    dbg('priority tasks', o, s='targets', l=3)
    return o


def force_number(i):
      try:
        o = float(i)
      except:
        o = 0
      return o


def get_times(task):
    day = start_of_day()
    week = day - timedelta(days = 7)
    month = day - timedelta(days = 30)

    o = {}
    task['parent_label'] = str(task['parent_label']) # This keeps causing a noneType + str concat error even though it should be a string already 

    o[task['label'] +' today'] = get_total_time(task['id'],'task',day)
    o[task['parent_label']+' today'] = get_total_time(task['parent_id'],'list',day)

    o[task['label'] +' weekly'] = get_total_time(task['id'],'task',week)
    o[task['parent_label']+' week'] = get_total_time(task['parent_id'],'list',week)    
    
    o[task['label'] +' monthly'] = get_total_time(task['id'],'task',month)
    o[task['parent_label']+' month'] = get_total_time(task['parent_id'],'list',month)

    for key, val in o.items():
        o[key] = sec_to_time(val)

    return o


def db_deactivate_todo(id):
    print("deactivating todolist",id)
    # Delete inactive todo items (except where there are sessions for them) 
    db_query("DELETE FROM tasks WHERE todolist = ? AND id NOT IN (SELECT task_id FROM sessions)",(id,))
    db_query("DELETE FROM lists WHERE todolist = ? AND id NOT IN (SELECT parent_id FROM tasks)",(id,))
    db_query("UPDATE lists SET status = -1 WHERE todolist = ?",(id,))
    db_query("UPDATE tasks SET status = -1 WHERE todolist = ?",(id,))


def get_todolists(use_db_cache = False):
    tasks = {}
    lists = {}  

    # TODO: Use threads to load todos concurrently?
    # with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
    #    executor.map(thread_function, range(3))
    
    for id, todo in conf.user['todolists'].items(): 
        if not todo['status']:
            db_deactivate_todo(id)
            continue # omit inactive connectors
        
        if use_db_cache:
            print('loading '+todo['type']+' todo '+id+" from db")

            todos = db_get_todolist(todo['id'])
        else:
            print('Refreshing '+todo['type']+' todo '+id+" ")

            try:
                todos = conf.todo_connectors[todo['type']].get_todos(todo)
                conf.todo_sync_time = now()
                db_set_todolist(todo['id'],todos['lists'],todos['tasks'])
            except Exception as e:
                # This will happen if offline

                # For testing
                if debug_level == 3 or 'todoloading' in debug_systems or todo['type'] in debug_systems :
                    raise e

                error_notice('Error Loading '+todo['label'],e)
                try:
                    todos = db_get_todolist(todo['id'])
                except Exception as e:
                    error_notice('Also Failed to load '+todo['label']+' From cache' ,e)
                    todos = {'lists': {}, 'tasks':{}}

        tasks.update(todos['tasks'])
        lists.update(todos['lists'])
    
    lists_cache(lists)

    o = {'lists': lists, 'tasks':tasks}
    return o

def get_most_recent_list(session = None):

    if session and 'parent_id' in session: 
       return  session['parent_id']

    last_session = db_query("SELECT parent_id FROM sessions WHERE parent_id IN (SELECT id FROM lists WHERE status = 1) ORDER BY sessions.start_time DESC LIMIT 1")
    if last_session: 
        return last_session[0]['parent_id']


def choose_from_lists(callback, selected_list_id = None, session = None, accepts_tasks = True):
    ''' Returns a Gtk.ScrolledWindow widget with radio button'''

    # get last used list
    if not selected_list_id:
        selected_list_id = get_most_recent_list(session = None)

    box = Gtk.VBox()

    scrolled_window = Gtk.ScrolledWindow()
    scrolled_window.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
    scrolled_window.set_propagate_natural_height(True)
    scrolled_window.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
    scrolled_window.set_size_request(-1, 400)
    # scrolled_window.set_min_content_height(500)
    scrolled_window.add(box)

    todolist = None

    where = " WHERE status = '1' "

    if accepts_tasks:
        where += " AND data like '%\"accepts_tasks\": true%' "

    lists = db_query("SELECT * FROM lists "+where+" ORDER BY todolist, extended_label ASC ", None, 'id')

    for id, l in lists.items():
        l['data'] = json.loads(l['data']) 
        # print("AddTask list ",l['label'])

        if todolist == None:
            # initial button group
            button_group = Gtk.RadioButton(label=l['label'])
            item = button_group
        else:
            item = Gtk.RadioButton(label=extended_label(l), group=button_group)

        if todolist != l['todolist']:
            # new buttongroup for new todolist
            todolist = l['todolist']
            l['header'] = Gtk.Label()
            l['header'].set_markup("<b>"+conf.user['todolists'][todolist]["label"]+"</b>")
            box.add(l['header'])
            
        if l['id'] == selected_list_id:
            item.set_active(True)

        item.connect("toggled", callback, l)
        box.add(item)

    return scrolled_window


def add_todos_to_menu(target_menu = None, menu_tasks = None, list_menus = None, activate_callback = None, use_db_cache = True):

    todos = get_todolists(use_db_cache)

    lists = todos['lists']  
    tasks = todos['tasks']

    # print('Add tasks to menu:')
    # print('refreshed lists',json.dumps(lists, indent=4))
    # print(json.dumps(tasks, indent=4))

    list_menu_items = {}
    priority_tasks = {}

    if conf.user['display_todolist_as_top_level_list'] == 'always':
        toplevel_todos = True
    elif conf.user['display_todolist_as_top_level_list'] == 'never':
        toplevel_todos = False
    else:
        # Count top level lists and nest if to many 
        toplevel_count = 0

        # items = lists
        items = lists | tasks
        for id, i in items.items():
            if i['status'] == 1 and i['parent_id'] == i['todolist']:
                toplevel_count += 1
            elif 'priority' in i and i['priority']:
                # count priority tasks
                toplevel_count += 1


        print('num top level lists',toplevel_count)
        dbg('toplevel_count',toplevel_count,s='menu')


        if toplevel_count > conf.user['max_top_level_menu_items']:
            toplevel_todos = True
        else:
            toplevel_todos = False
    
    # for testing
    # toplevel_todos = True
    # toplevel_todos = False
    
    # separator after priority tasks
    target_menu.prepend(Gtk.SeparatorMenuItem.new())

    # Create list sub menus 
    for list_id, l in lists.items():
        if l['status'] != 1:
            continue

        if not toplevel_todos and l['id'] == l['todolist']:
            continue

        list_menus[list_id] = Gtk.Menu() # the sub_list that items get added to
        list_menu_items[list_id] = Gtk.MenuItem.new_with_label(lists[list_id]['label']) # the "item" that gets added 
        list_menu_items[list_id].set_submenu(list_menus[list_id])

    # Add sub menus (items) to parents (sub_menus)
    for list_id, l in lists.items():
        if l['status'] == 1:

            if not toplevel_todos and l['id'] == l['todolist']:
                continue
            
            if 'priority' in l and l['priority']:
                target_menu.prepend(list_menu_items[list_id])
            elif (toplevel_todos and l['id'] == l['todolist']) or (not toplevel_todos and l['parent_id'] == l['todolist']):
                target_menu.append(list_menu_items[list_id])
            else:
                list_menus[l['parent_id']].append(list_menu_items[list_id])



    for id, t in tasks.items():
        if t['status'] == 1:

            menu_tasks[t['id']] = Gtk.MenuItem.new_with_label(str(t['label']))
            menu_tasks[t['id']].connect("activate", activate_callback, t)


            if 'priority' in t and t['priority'] > 0 and ('priority' not in lists[t['parent_id']] or lists[t['parent_id']]['priority'] in [0,None]):
                # Hoist priority tasks, in correct order
                priority_tasks[t['id']] = t['priority']

            elif not t['parent_id'] or t['parent_id'] not in list_menus:
                target_menu.append(menu_tasks[t['id']])
            else:
                # print("add "+t['label']+" to sub menu "+t['parent_label'] )
                list_menus[t['parent_id']].append(menu_tasks[t['id']])


    # Sort and add priority tasks
    if priority_tasks:

        priority_tasks = dict(sorted(priority_tasks.items(), key=lambda item: item[1], reverse=True))
        for t_id in priority_tasks:
            target_menu.prepend(menu_tasks[t_id])

    
    # Ugly (but effectively?) remove empty lists
    for key, menuitem in list_menu_items.items():
        if len(menuitem.get_children()) == 0:
            print("remove empty list", lists[key]['label'])
            menuitem.destroy()


    # Do it again to remove upper level empties!
    for key, menuitem in list_menu_items.items():
        if len(menuitem.get_children()) == 0:
            print("remove empty list", lists[key]['label'])
            menuitem.destroy()

    target_menu.show_all()

def datetime_minus_calendar_unit(unit = 'days', num = 1, ref_date = None):
    ''' returns a datetime for the start of the unit num units ago (from the optional ref_date) 
    unit can be days, weeks, months, years 
    num  
    ''' 
    # Looks weird but...
    num = num - 1 
    if not ref_date:
        ref_date = start_of_day()
    if unit in ['day','days']:
        o = ref_date - timedelta(days=num) 
        if num > 0: o = o - timedelta(days=num)
    if unit in ['week','weeks']:
        o = ref_date - timedelta(days=ref_date.weekday())
        if num > 0: o = o + relativedelta(dt1= o, weeks= - num)
    if unit in ['month','months']:
        o = ref_date.replace(day=1)
        if num > 0: o = o + relativedelta(dt1= o, months= - num)
    if unit in ['year','years']:
        o = ref_date.replace(month=1,day=1)
        if num > 0: o = o + relativedelta(dt1= o, years= - num)
    return o

def get_time_target_start_time(t):
    ''' t must be a time target dict from conf.user'''



def check_time_target(i,item_type = 'tasks', min_max_type = None):

    ''' No target returns False 
     max or min returns float of target minus sessions ''' 

    # Backward compatibility 
    if item_type == 'task': item_type = 'tasks'
    if item_type == 'list': item_type = 'lists'

    ballance = False

    if i['id'] in conf.user['time_targets'][item_type]:
        t = conf.user['time_targets'][item_type][i['id']]
    elif i['parent_id'] in conf.user['time_targets']['lists']:
        # TODO: make this infinitely recursive using get lists_for_task
        t = conf.user['time_targets']['lists'][i['parent_id']]
    else:
        dbg(str(i['label']) + " check_time_target no applicable target:", s='targets', l=3 )
        return False
        
    if not min_max_type or t['type'] == min_max_type:
        target_start = datetime_minus_calendar_unit(t['within_unit'],t['within_value'])

        ballance = t['value'] - get_total_time(i['id'],item_type, target_start, None, "get_minutes")

        dbg(i['label'] + " check_time_target ballance:" + str(ballance), s='targets', l=2 )
        return ballance
        
    dbg(i['label'] + min_max_type + " check_time_target no applicable min_max_type target:", s='targets', l=3 )
    return False



import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, Gdk
import conf
import utils
from utils import *


class SessionOptionsDialog(Gtk.Dialog):

    def __init__(self, parent=None, input_data = None):

        
        if isinstance(parent,Gtk.Window):
            self.parent_window = parent
            if isinstance(parent,Gtk.Application):
                self.app = parent.app
            else:
                error_notice("Error: NewTaskWDialog called out of application context")

        elif isinstance(parent,Gtk.Application):
            # if called from main class
            self.parent_window = None
            self.app = parent
        else:
            error_notice("Error: NewTaskWDialog called out of application context")

        session = self.app.session

        super().__init__(title="Modify Current Session", transient_for=self.parent_window, flags=0)
        self.set_position(position=1) # doesn't work on wayland
        self.set_default_size(300,300)
        self.set_border_width(15)

        box = self.get_content_area()

        print(' SessionOptionsDialog input_data',input_data)

        # label = Gtk.Label()
        # label.set_markup("<b>"+session['extended_label']+"</b> ")
        # box.add(label)

        menubar = Gtk.MenuBar()
        self.menu_bar_item = Gtk.MenuItem(label=session['label']+" (change)")
        menubar.append(self.menu_bar_item)
        change_task_menu = Gtk.Menu()
        self.menu_bar_item.set_submenu(change_task_menu)

        utils.add_todos_to_menu(change_task_menu, {}, {}, self.change_task,True)
        
        box.add(menubar)

        box.add(Gtk.Box(border_width=10)) #spacer

        duration_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=5)
        duration_box.add(Gtk.Label(label='Duration (in Minutes)'))
        
        adjustment = Gtk.Adjustment(value=(utils.force_number(session['duration']) / 60),
            lower=0,
            upper=999,
            step_increment=1,
            page_increment=5,
            page_size=0)

        self.duration_input = Gtk.SpinButton(adjustment=adjustment) 

        duration_box.add(self.duration_input)

        box.add(duration_box) #spacer
        box.add(Gtk.Box(border_width=10)) #spacer

        box.add(Gtk.Label(label='Notes'))

        self.notes_text_buffer = Gtk.TextBuffer()
        if 'notes' in session:
            self.notes_text_buffer.set_text(session['notes'])
        else:
            self.notes_text_buffer.set_text('\n\n') # hack to give it some hight

        self.notes = Gtk.TextView(buffer=self.notes_text_buffer)
        box.add(self.notes)

        box.add(Gtk.Box(border_width=10)) #spacer

        self.mark_done = Gtk.CheckButton(label="Mark Done")

        box.add(self.mark_done)

        self.add_button("Cancel", 1)

        self.add_button("Modify", 2)

        self.add_button("Finish", 3)
                
        self.set_default_response(3)

        self.show_all()
        self.connect("response", self.on_response)


    def change_task(self, widget = None, t = None):
        session = self.app.session

        print('Hot-swap current task (initial) session',session)
        session['label'] = t['label']
        session['extended_label'] = extended_label(t)
        session['id'] = t['id']
        session['parent_id'] = t['parent_id']
        session['parent_label'] = t['parent_label']
        session['task'] = t
        print('updated session',session)

        self.menu_bar_item.set_label(session['label'])


    def on_response(self, widget, response):
        
        session = self.app.session

        print("on_response",response, "original session", session)


        if response == 1:
            self.app.stop_task(None,'cancel')
            
        elif response > 1:

            if self.notes_text_buffer.get_modified():
            
                start, end = self.notes_text_buffer.get_bounds()
                session['notes'] = self.notes_text_buffer.get_text(start, end,False)

            print("(Modified) session", session)

            if response == 2:
                # move session [start_time] to set duration for 'continue' option 
                print('initial start_time', session['start_time'],'timedelta')

                session['start_time'] = now() - timedelta(seconds = (self.duration_input.get_value_as_int() * 60))

                print(' start_time + ', session['start_time'])

            # Or change session[duration] value  to reflect dropped / afk session 
            if response == 3:
                # Create custom end time
                custom_end_time = session['start_time'] + timedelta(seconds = (self.duration_input.get_value_as_int() * 60))

                # session['duration'] = (self.duration_input.get_value_as_int() * 60)
                
                if self.mark_done.get_active():
                    self.app.stop_task(None,'mark_done',custom_end_time)
                else:
                    self.app.stop_task(None,'save',custom_end_time)

        del self.app.session_options_dialog
        self.destroy()

















# sudo apt install gir1.2-appindicator3-0.1 

import os
import subprocess, sys
import signal
import json
import time
from datetime import datetime, timezone, timedelta
import importlib
from urllib.request import urlopen
import requests
import threading
import copy
from playsound3 import playsound
import setproctitle

import argparse 

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('AppIndicator3', '0.1')
gi.require_version('Notify', '0.7')

from gi.repository import Notify as notify
from gi.repository import Gtk, GLib, Gdk, Gio
# from gi.repository import GObject as gobject
from gi.repository import AppIndicator3 as appindicator

from dbus_idle import IdleMonitor


# Set working dir to file location 
os.chdir(os.path.dirname(os.path.realpath(__file__)))


print('sys.path',sys.path)
# Add working dir to path
sys.path.append(os.path.dirname(__file__))

print('sys.path with os.path.dirname(__file__) added',sys.path)


# from . import conf # this works in module context but not running as pile-of-files 
import conf # this works running as pile-of-files but not in module context 

import utils
from utils import *

from settings import SettingsWindow
from task_window import TaskWindow
from new_task_dialog import NewTaskWDialog
from session_options import SessionOptionsDialog 

dbg(conf.user,l=3,s='user_settings')




setproctitle.setproctitle(conf.app_name)
# notify.init(conf.app_name)


# TODO: call with arguments
# parser = argparse.ArgumentParser()
# parser.add_argument('packages', nargs='*')


class Application(Gtk.Application):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, application_id="org.example.myapp", **kwargs)
        
        self.window = None

        self.is_running = False
        self.session = {
            "label":"Randomness",
            "extended_label": 'Randomness',
            "start_time":datetime.now(),
            # "start_time":datetime.now(timezone.utc),
            "duration":0,
            'task':{},
            'notes':'',
        }

        self.menu_tasks = {}
        self.list_menus = {}

        self.indicator = appindicator.Indicator.new(conf.app_name, os.path.abspath('icon/icon-0.svg'), appindicator.IndicatorCategory.APPLICATION_STATUS)

        self.indicator.set_status(appindicator.IndicatorStatus.ACTIVE)

        self.menu = Gtk.Menu()
        # menu.set_reserve_toggle_size(False) # skip menu left padding,  doesn't work 


        utils.db_init()
        utils.db_update()

        # db_cleanup()

        # self.update_menu()
        self.async_refresh()

        self.indicator.set_menu(self.menu)

        main_tick_timer = GLib.timeout_add_seconds(conf.user['tick_interval'], self.tick)

        try:
            db_session = db_query("SELECT value FROM system WHERE field = 'session'")

            if db_session:
                s = json.loads(db_session[0]['value'])
                s['start_time'] = datetime.strptime(s['start_time'],'%Y-%m-%d %H:%M:%S.%f')
                self.session = s
                self.is_running = True
                dbg("resuming session",s['label'],l=2, s='session')
        except Exception as e:
            dbg("Error resuming session",e,l=1)


        # Testing
        # time.sleep(3)
        # self.open_session_options_dialog('test_param')

        # self.open_settings_window() #for testing
        # self.open_task_window() #for testing

        # time.sleep(2)
        # self.open_new_task_dialog() #for testing

        # time.sleep(2)
        # self.print_time_totals()

        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGUSR1, self.signal_handler)
        signal.signal(signal.SIGUSR2, self.signal_handler)

        Gtk.main()


    def print_time_totals(self):

        # SELECT extended_label, (SUM(duration) / 60 ) FROM sessions GROUP BY extended_label
        # SELECT extended_label, (SUM(duration) / 60 ) FROM sessions GROUP BY parent_id
        l = ''
        for t in db_query("SELECT DISTINCT id, parent_id, label, parent_label FROM tasks WHERE id IN(SELECT DISTINCT task_id FROM sessions) ORDER BY parent_label ASC "):
            if t['parent_id'] and l != t['parent_id']:
                l = t['parent_id']
                print(extended_label(db_get_item_by_id(t['parent_id'],'lists')) +": "+ sec_to_time(get_total_time(l,'lists')))

            print("   "+ t['label']  +": "+  sec_to_time(get_total_time(t['id'])))
            # def get_total_time(id, category = 'tasks', start_time = None, end_time = None, get_minutes = None):


    # def do_command_line(self, command_line):
    #     # Handle command-line arguments
    #     options = command_line.get_options_dict()
    #     print('do_comand_line options', options)


    def quit(self, w):
        notify.uninit()
        Gtk.main_quit()


    def check_afk_time(self):
        # returns seconds of inactivity
        # See https://stackoverflow.com/questions/67083083/how-to-get-idle-time-in-linux  or 

        # Works on x11 but not wayland
        # Requires xprintidle (sudo apt install xprintidle)
        # idle_time = int(int(subprocess.getoutput('xprintidle')) / 1000)

        # Currently using: https://github.com/bkbilly/dbus_idle
        # Requires: 
        #   sudo apt install meson libdbus-glib-1-dev patchelf
        #   pip install dbus-idle

        idle_time = int(int(IdleMonitor().get_dbus_idle()) / 1000)
        return idle_time


    def tock(self):

        afk_time = self.check_afk_time()
        # print("Idle time: "+str(afk_time))

        dbg('Last todo todo_sync_time', conf.todo_sync_time, 'Time diff',int(time_difference(conf.todo_sync_time)),'Auto refresh interval * 60', (conf.user['todolist_refresh_interval'] * 60), s="todoloading", l=3 )

        if (int(time_difference(conf.todo_sync_time)) / 60) > conf.user['todolist_refresh_interval'] * 60 :
            self.async_refresh()

        minutes = (int(self.session['duration']) / 60)
        
        if(self.is_running == False):

            if float(minutes / conf.user['randomness_interrupt_interval']).is_integer():
                if afk_time < 30:
                    # if self.taskwindow:
                        # self.taskwindow.present()
                        # self.taskwindow.the_question = Gtk.Label("Your randomness timer is at "+str(minutes)+" minutes. ")
                    # else:
                    
                    notify.Notification.new("What Am I doing?","Your randomness timer is at "+str(minutes)+" minutes. ", None).show()

                    self.open_task_window()
                        
                    playsound('sound/dinner-bell.mp3',False)

                elif afk_time > 120:
                    self.session['duration'] = 0
                    self.session['start_time'] = now()
                    print("Idle time reset. afk:",  afk_time, self.session)

        else:
            if afk_time > 120:
                # only show this once
                if afk_time < 181:
                    self.open_task_window(None,{'afk_time':afk_time})
                    # session_options_dialog(None, 'test input_data')
            else:
                if self.session['label'] in conf.user['custom_pomodoro_intervals']:
                    check = conf.user['custom_pomodoro_intervals'][self.session['label']]
                else:
                    check = conf.user['pomodoro_interval']
                
                if float(minutes / check ).is_integer():
                    notify.Notification.new("Time for a Break?","You've been working on "+self.session['label']+" for "+str(minutes)+" minutes. ", None).show()

                    playsound('sound/bell-xylophone-g.mp3',False)


                if 'target' in self.session:
                    t = self.session['target']

                    t['percent'] = ( (t['starting_value'] + minutes) / t['value'] ) * 100
                    print('target',t)

                    if t['percent'] >= 100:
                        if t['type'] == 'max':

                            notify.Notification.new("Time is up for this "+t['scope'],"You'r at "+str(round(t['percent']))+"% of your "+str(t['value'])+" minutes in the last "+str(t['within_value'])+" "+ t['within_unit'], None).show()

                            playsound('sound/dinner-bell.mp3',False)


                    # maybe add a target % to the session and show with tick

    icon_tick_number = 0

    def tick(self):
        menu = self.menu    
        indicator = self.indicator

        self.session['duration'] = int(time_difference(self.session['start_time']))

        if(self.session['duration'] > 2 and (int(self.session['duration']) / 60).is_integer()):
            self.tock()

        if(self.is_running == True):
            self.icon_tick_number = self.icon_tick_number + 1
            
            if self.icon_tick_number > 9:
                self.icon_tick_number = 1    

            label = self.session['label'] + ": " + sec_to_time(self.session['duration'])

            indicator.set_icon_full(os.path.abspath('icon/icon-'+str(self.icon_tick_number)+'.svg'),label) # Proper but doesn't work on mint
            # indicator.set_icon(os.path.abspath('icon/icon-'+str(self.icon_tick_number)+'.svg')) # depreciated but working
        
        else:

            # label = random.choice(conf.idle_messages) # Cool but makes menu bounce around #Could be paused when the menu opens
            label = conf.user['default_text']

            indicator.set_icon_full(os.path.abspath('icon/icon-0.svg'),label)
            # indicator.set_icon(os.path.abspath('icon/icon-0.svg')) # depreciated but working


        # https://lazka.github.io/pgi-docs/#AyatanaAppIndicator3-0.1/classes/Indicator.html#AyatanaAppIndicator3.Indicator.set_label
        indicator.set_label(label, "Wide")
        return True


    def start_task(self, w = '',task_data = None):

        task_label = task_data['label']
        dbg("starting "+ task_label,l=1)

        if(self.is_running == True):
            self.stop_task()
        
        self.is_running = True
        
        s = {
            "label": task_label,
            "extended_label": utils.extended_label(task_data),
            "start_time": now(),
            "duration":0,
            "task":task_data,
            "notes":'',
        }

        if s['task']['id'] in conf.user['time_targets']['tasks']:
            s['target'] = copy.copy(conf.user['time_targets']['tasks'][s['task']['id']])
            s['target']['scope'] = 'task' 
            target_start = datetime_minus_calendar_unit(s['target']['within_unit'],s['target']['within_value'])

            s['target']['starting_value'] = utils.get_total_time(task_data['id'],'task', target_start,None,"get_minutes")
            s['target']['percent'] = s['target']['starting_value'] / s['target']['value'] * 100

        if s['task']['parent_id'] in conf.user['time_targets']['lists']:
            s['target'] = copy.copy(conf.user['time_targets']['lists'][s['task']['parent_id']])
            s['target']['scope'] = 'list' 
            target_start = datetime_minus_calendar_unit(s['target']['within_unit'],s['target']['within_value'])

            s['target']['starting_value'] = utils.get_total_time(task_data['parent_id'],'list', target_start,None,"get_minutes")
            s['target']['percent'] = s['target']['starting_value'] / s['target']['value'] * 100
            
            dbg("session with list target started",s,s="targets",l=3)

        self.session = s

        # TODO: copy this to save change in session 
        db_session = copy.deepcopy(s)
        db_session['start_time'] = db_session['start_time'].strftime("%Y-%m-%d %H:%M:%S.%f%z")
        db_query("REPLACE INTO system(field, value) VALUES(:field, :value)", {'field':'session','value':json.dumps(db_session)})

        self.tick()

        # print('total times for task ',task_label, utils.get_times(task_data))    

        self.task_running_menu_additions()

        self.menu.show_all()


    def stop_task(self, w = '', action = 'save', custom_end_time=None):

        list_menus = self.list_menus
        menu_tasks = self.menu_tasks
        session = self.session
        
        print("Stopping ", session['label'])
        task = session['task']
        if self.is_running == True:
            session['duration'] = time_difference(session['start_time'],custom_end_time)
            db_query("DELETE FROM system WHERE field = 'session' ")

            if action != 'cancel':

                # Get time tracker for this tasks todolist 
                todolist_conf = conf.user['todolists'][task['todolist']]
                timetracker_conf = conf.user['timetrackers'][todolist_conf['timetracker']]

                dbg("Save Session to "+ todolist_conf['timetracker'])
                try:
                    save_thread = threading.Thread(target=conf.timetracker_connectors[timetracker_conf['type']].save_session, args=(session,timetracker_conf) )
                    save_thread.start()
                except Exception as e:
                    error_notice('Error Saving Time Data'," Recording timetracking for "+ task['label']+"  in "+timetracker_conf['label']+" had a serious failure",e )

                session['timetracker'] = todolist_conf['timetracker']

                utils.db_save_session(session)

                if action == 'mark_done':

                    try:
                        done_thread = threading.Thread(target=conf.todo_connectors[todolist_conf['type']].mark_task_done, args=(task,) )
                        done_thread.start()
                    except Exception as e:
                        error_notice('Error Marking Task Done'," Marking "+ task['label']+" as done in "+todolist_conf['label']+" had a serious failure",e )

                    db_query("UPDATE tasks set status = '0' WHERE id = ? ",(task['id'],) )

                    print('remove menu item')
                    self.menu_tasks[task['id']].destroy()
                    playsound('sound/xylophone-chord.mp3',False)

                else: 
                    print("add ",utils.extended_label(task)," to recent tasks")
                    i = Gtk.MenuItem.new_with_label(utils.extended_label(task))
                    i.connect('activate',self.start_task,task)
                    list_menus['recent'].prepend(i)
                    try:
                        # comment: 
                        list_menus['recent'].get_children()[11].destroy()
                    except Exception as e:
                        dbg("Exception trying to rotate recent tasks. probably are less than 11",l=2,s="recent")
                        # end try

                    # Check de-hoist completed time targets
                    ballance = utils.check_time_target(task)
                    dbg('time target ballance:', ballance,l=2,s='time_targets')

                    if ballance != False and ballance < 0:
                        dbg('un-hoist completed time target',l=2,s='time_targets')
                        try:
                            self.menu.remove(menu_tasks[task['id']]) # how to check that this actually worked/ it was hoisted?
                            list_menus[task['parent_id']].append(menu_tasks[task['id']])
                        except Exception as e:
                            # Boo Hoo?
                            dbg('de-hoisting failed',e)
                        playsound('sound/xylophone-chord.mp3',False)

                    self.menu.show_all()

            self.is_running = False
            # print(utils.get_times(task))
        
            notify.Notification.new(action.capitalize()+" "+session['label']+" "+sec_to_time(session['duration']),utils.pretty_dict(utils.get_times(task)), None).show()
        
            # Start randomness timer
            self.session = {
                "label": 'Randomness',
                "extended_label": 'Randomness',
                "start_time": now(),
                "duration":0,
                "task":{}
            }

            self.tick()

            # Rm stop task menu item
            self.menu.get_children()[0].destroy()
            self.menu.get_children()[0].destroy()

        else:
            print('no task running!')


    def task_running_menu_additions(self):
        
        i = Gtk.ImageMenuItem.new_with_label("Edit Session")
        i.set_image(Gtk.Image.new_from_file(os.path.abspath('icon/edit.svg'))) 

        # i.set_image(Gtk.Image.new_from_file('/usr/share/icons/Yaru/scalable/status/software-installed-symbolic.svg')) 
        i.set_always_show_image(True) 
        i.connect("activate", self.open_session_options_dialog,'from_menu')
        self.menu.insert(i,0)    

        i = Gtk.ImageMenuItem.new_with_label("Pause" )
        i.set_image(Gtk.Image.new_from_file(os.path.abspath('icon/pause.svg'))) 

        # i.set_image(Gtk.Image.new_from_file('/usr/share/icons/Yaru/scalable/multimedia/pause-symbolic.svg')) 
        i.set_always_show_image(True) 
        i.connect("activate", self.stop_task)
        self.menu.insert(i,0)


    def async_refresh(self, w=None):

        self.indicator.set_label("Refreshing Todolists", "Wide")
        menu_item = Gtk.MenuItem.new_with_label("Refreshing Todolists")
        # setting_menu_item.connect("activate",self.open_settings_window) # perhaps give the option to abort?
        self.menu.append(menu_item) 
        self.menu.show_all()

        connectors_thread = threading.Thread(target=self.async_refresh_inner)
        connectors_thread.start()


    def async_refresh_inner(self):
        # dbg("async refresh started",s="todoloading",l=3)
        utils.get_todolists()
        dbg("async refresh complete",s="todoloading",l=3)
        GLib.idle_add(self.update_menu)
        
        try:
            GLib.idle_add(self.taskwindow.update_default_tasks)
        except Exception as e:
            # raise e
            dbg("exception refreshing taskwindow default tasks, it probably doesn't exist",l=3,s="taskwindow") 


    def update_menu(self, w = ''):
        # print("update_menu")
        menu = self.menu

        self.menu_tasks = {}
        self.list_menus = {}

        menu_tasks = self.menu_tasks
        list_menus = self.list_menus


        menu.foreach(lambda child: child.destroy()) 

        utils.add_todos_to_menu(menu, menu_tasks, list_menus, self.start_task)

        menu.append(Gtk.SeparatorMenuItem.new())

        # Recent Tasks Menu
        list_menus['recent'] = Gtk.Menu() # the sub_list that items get added to
        recents = Gtk.MenuItem.new_with_mnemonic("Recent Tasks") # the "item" that gets added 
        recents.set_submenu(list_menus['recent'])
        menu.append(recents)

        for id, t in utils.get_recent_tasks().items():
            # print('recent task',t)
            try:
                i = Gtk.MenuItem.new_with_label(utils.extended_label(t))
                i.connect('activate',self.start_task,t)
                list_menus['recent'].append(i)

            except Exception as e:
                # Because a key error if the task has was completed or todolist removed
                print('recent_tasks error', e)
        menu.append(Gtk.SeparatorMenuItem.new())


        # Todolist openables
        openables = []
        for id, todolist in conf.user['todolists'].items():
            if todolist['status']:
                openable = get_connector_openable(None, todolist,False)
                if openable not in openables:            
                    openables.append(openable)
                    
                    menu_aw = Gtk.MenuItem.new_with_label("Open "+todolist['label']+" ")
                    menu_aw.connect("activate", get_connector_openable, todolist)
                    menu.append(menu_aw)
            
        # Timetrackers
        for id, timetracker in conf.user['timetrackers'].items():
            if timetracker['status']:
                openable = get_connector_openable(None, timetracker,False)
                if openable not in openables:            
                    openables.append(openable)
                    menu_aw = Gtk.MenuItem.new_with_label("Open "+timetracker['label']+" ")
                    menu_aw.connect("activate", get_connector_openable, timetracker)
                    menu.append(menu_aw)

        # Update List
        # Lots of spaces to make menu wide-ish in preparation for opening wide sub menus 
        menu_update = Gtk.MenuItem.new_with_label("Refresh Lists                                                                    ")
        menu_update.connect("activate", self.async_refresh)
        menu.append(menu_update)

        setting_menu_item = Gtk.MenuItem.new_with_label("Settings")
        setting_menu_item.connect("activate",self.open_settings_window)
        menu.append(setting_menu_item)    
        
        taskswindow_menu_item = Gtk.MenuItem.new_with_label("Tasks")
        taskswindow_menu_item.connect("activate",self.open_task_window)
        menu.prepend(taskswindow_menu_item)


        new_task_menu_item = Gtk.MenuItem.new_with_label("New Task")
        new_task_menu_item.connect("activate",self.open_new_task_dialog)
        menu.append(new_task_menu_item)
        
        quit_menu_item = Gtk.MenuItem.new_with_label("Quit")
        quit_menu_item.connect("activate",self.quit)
        menu.append(quit_menu_item)
        
        if self.is_running:
            self.task_running_menu_additions()

        menu.show_all()


    def open_session_options_dialog(self, w = None, params = None):

        if hasattr(self, 'session_options_dialog'):
            # print("present existing session_options_dialog",self.session_options_dialog)
            self.session_options_dialog.present()
        else:
            self.session_options_dialog = SessionOptionsDialog(self,params)
            self.session_options_dialog.show_all()


    def open_settings_window(self, w = None, **kwargs):

        if hasattr(self, 'settings_window'):
            self.settings_window.present()
        else:
            self.settings_window = SettingsWindow(self,**kwargs)
            self.settings_window.show_all()


    def open_task_window(self, w = None, data = None):
        self.taskwindow = TaskWindow(self, data)


    def open_new_task_dialog(self, w = None, passed_data = None):

        if hasattr(self, 'new_task_dialog'):
            self.new_task_dialog.present()
        else:
            self.new_task_dialog = NewTaskWDialog(self,passed_data)
            self.new_task_dialog.show_all()


    def signal_handler(self, sig, frame):
        dbg('Signal received',sig,s='signals')

        if sig == signal.SIGUSR1:
            self.open_task_window()
        
        elif sig == signal.SIGUSR2:
            self.open_session_options_dialog()

        else: 
            dbg("no handler for received signal",s='signals',l=3)


def main_entrypoint():
    signal.signal(signal.SIGUSR1, Application.signal_handler) 
    app = Application()

if __name__ == "__main__":
    signal.signal(signal.SIGUSR1, Application.signal_handler) # almost works
    app = Application()






import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, Gdk
import conf
import utils
from utils import *

class NewTaskWDialog(Gtk.Dialog):

    def __init__(self, parent=None, passed_data=None):

        if isinstance(parent,Gtk.Window):
            self.parent_window = parent
            if isinstance(parent,Gtk.Application):
                self.app = parent.app
            else:
                error_notice("Error: NewTaskWDialog called out of application context")

        elif isinstance(parent,Gtk.Application):
            # if called from main class
            self.parent_window = None
            self.app = parent
        else:
            error_notice("Error: NewTaskWDialog called out of application context")

        super().__init__(title="New Task ", transient_for=self.parent_window, flags=0)

        self.present() # try to focus the window
        self.set_position(position=1) # center dialog, doesn't work on wayland
        self.set_default_size(300,300)

        self.set_border_width(15)

        box = self.get_content_area()

        self.task_label_entry = Gtk.Entry()
        self.task_label_entry.set_placeholder_text('Task')

        if passed_data and 'label' in passed_data:
            self.task_label_entry.set_text(passed_data['label'])

        box.add(self.task_label_entry)

        box.add(Gtk.Box(border_width=10)) #spacer

        default_list = get_most_recent_list(self.app.session)
        if default_list:
            self.selected_list = db_get_item_by_id(default_list,table='lists')

        box.add(choose_from_lists(self.select_list, default_list))

        box.add(Gtk.Box(border_width=10)) #spacer
        
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL
        )

        self.add_button("Save", 1)
        
        self.add_button("Save & Start", 2)
        
        self.set_default_response(2)

        self.show_all()
        self.connect("response", self.on_response)


    def select_list(self, widget = None, l = None):
        if widget.get_active():
            self.selected_list = l


    def on_response(self, widget, response):
        list_menus =self.app.list_menus
        menu_tasks = self.app.menu_tasks
        # print("Cool!",response, self.selected_list,self.task_label_entry)

        if response == 1 or response == 2:

            task_label = self.task_label_entry.get_text()

            if not task_label:
                error_notice("Please Enter a name before saving new task")
                return False

            parent_list = self.selected_list
            todolist_conf = conf.user['todolists'][parent_list['todolist']]
            try:
                task = conf.todo_connectors[todolist_conf['type']].add_new_task(todolist_conf,parent_list,task_label)
            
                # add it to the database 
                db_set_item(task)

                # add it to the menu 
                menu_tasks[task['id']] = Gtk.MenuItem.new_with_label(task['label'])
                menu_tasks[task['id']].connect("activate", self.app.start_task, task)
                list_menus[task['parent_id']].append(menu_tasks[task['id']])

                if response == 2:
                    self.app.start_task(None, task)

                    try:
                        self.app.taskwindow.destroy()
                    except Exception as e:
                        print("Error trying to remove taskwindow",e)
                    
                else:
                    try:
                        self.app.taskwindow.task_search()
                    except Exception as e:
                        print("Error trying to update taskwindow",e)

                    
            except Exception as e:
                #   for testing

                if utils.debug_level > 1:
                    raise e
                
                error_notice('Error adding tasks',"Adding "+ task_label+" to "+todolist_conf['label']+" had a serious failure",e )

            print('connector add task response',task)
        elif response == Gtk.ResponseType.CANCEL:
            print("Cancel button clicked")
 
        del self.app.new_task_dialog
        self.destroy()




import json
import time
from datetime import datetime, timezone

from urllib.request import urlopen

import requests

import conf
import utils

# UPDATE sessions SET task_id = "Vikunja_t" || task_id WHERE todolist = 'Vikunja';
# UPDATE sessions SET parent_id = "Vikunja_l" || parent_id WHERE todolist = 'Vikunja';

def vikunja_item_id(id, item_type, user_conf):
    return user_conf['id'] +'_'+ item_type[0] + str(id)

def add_new_task(user_conf,parent_list,task_label):
    head = {'Authorization': 'Bearer {}'.format(user_conf['token'])}
    headers = {'content-type': 'application/json'}
    data = {'title': task_label}

    utils.dbg('add_new_task() parent_list',parent_list, l=0)

    response = requests.put(user_conf['url']+"api/v1/projects/"+str(parent_list['data']['id'])+"/tasks",json=data, headers=head)

    task_data = response.json()

    print("Add task "+task_label+" to "+str(parent_list['id'])+" Response",task_data)

    id = vikunja_item_id(task_data['id'], 'task', user_conf)

    task = {
            'id':id,
            'label':task_data['title'],
            'parent_id':parent_list['id'],
            'parent_label':parent_list['label'],
            'todolist':user_conf['id'],
            'data':task_data
        }

    return task


def mark_task_done(task):
    user_conf = conf.user['todolists'][task['todolist']]
   
    head = {'Authorization': 'Bearer {}'.format(user_conf['token'])}
    headers = {'content-type': 'application/json'}
    data = {'done': True,}
    response = requests.post(user_conf['url']+"api/v1/tasks/"+str(task['data']['id']),json=data, headers=head)

    # print(response.json()['done'])
    return response.json()['done']
    
    # print('\nresponse.url')
    # print(response.url)

    # print('\nresponse.status_code')
    # print(response.status_code)

    # print('\nrequest.body')
    # print(response.request.body)

    # print('\nresponse')
    # print(response.json())


# Token auth for this doesn't seem to work propery, boo, get it from the user  

# def get_vikunja_user_data(user_conf,user_data_cache = {}):
#     ''' get (and cache) vikunja user data from API 
#     '''
#     # See https://try.vikunja.io/api/v1/docs#tag/user/paths/~1user/get

#     if not user_data_cache:
#         # try:
#         headers = {"Authorization": "Bearer "+user_conf['token']}
#         print("get_vikunja_user_data headers")
#         print(headers)
#         user_data_cache = requests.get(user_conf['url']+"api/v1/user", headers=headers).json()
#         # except Exception as e: 
#             # error_notice("Vikunja Error",e)
#     print("get_vikunja_user_data")
#     print(user_data_cache)
#     return user_data_cache

    
def get_todos(user_conf):

    tasks = {}
    lists = {}

    lists[user_conf['id']] = {
            'id':user_conf['id'],
            'label':user_conf['label'],
            'parent_id':'',
            'parent_label':'',
            'todolist':user_conf['id'],
            'status': 1,
            'data':{
                'accepts_tasks':False
            }
        }
    

    priorities = {'DO NOW':1,'Urgent':2,'high':3}
    headers = {"Authorization": "Bearer "+user_conf['token']}
    projects_lists = requests.get(user_conf['url']+"api/v1/projects", headers=headers).json()

    # print("vikunja projects")
    # print(json.dumps(projects_lists, indent=4))

    # for i in projects_list:
    for i, itemIter in enumerate(projects_lists):

        p = projects_lists[i]

        #  Omit filter lists 
        if p['id'] < 0:
            continue 


        # print(p)
        # p['id'] = str(p['id'])
        id = vikunja_item_id(p['id'] , 'list', user_conf)

        # views are bulky and useless so remove them
        del p['views']
        
        lists[id] = {
            'id':id,
            'label':p['title'],
            'parent_id':user_conf['id'],
            'parent_label':user_conf['label'],
            'todolist':user_conf['id'],
            'data':p
        }
        
        lists[id]['data']['accepts_tasks'] = True

        if p['is_archived']: 
            lists[id]['status'] = -1
        else:
            lists[id]['status'] = 1

        if p['parent_project_id'] != 0: 
            parent_id = vikunja_item_id(p['parent_project_id'], 'list', user_conf)
            lists[id]['parent_id'] = parent_id


            if parent_id in lists:
                lists[id]['parent_label'] = lists[parent_id]['label']


        # NOTE: Originally tasks where queried per list, however this no longer works well because it requires a 'view' id, which comes wih various un-predictable baked-in filters (like done=true). Using the tasks/all api endpoint avoids this but quickly hits the default maxitemsperpage limit (which could, less obviously, show up in the per list method). 

        # else: print('parent_project_id', p['parent_project_id'], 'not in projects_lists')

        # view_id = p['views'][0]['id']
        # utils.dbg({'view_id':view_id},'vikunja')
        

        # Including 'done' tasks (Doesn't seem o includ them actually)'
        # projects_tasks = requests.get(user_conf['url']+"api/v1/projects/"+str(p['id'])+"/views/"+str(view_id)+"/tasks?sort_by[]=position&order_by[]=asc&equalspage=1", headers=headers).json()
        
        # Filtering out done tasks, Works
        # projects_tasks = requests.get(user_conf['url']+"api/v1/projects/"+str(p['id'])+"/views/"+str(view_id)+"/tasks?sort_by[]=position&order_by[]=asc&filter_by[]=done&filter_value[]=false&filter_comparator[]=equals&filter_comparator[]=equalspage=1", headers=headers).json()
        
        # projects_tasks = requests.get(user_conf['url']+"api/v1/projects/"+str(p['id'])+"/tasks", headers=headers).json()

        # print("vikunja projects_tasks")
        # print(json.dumps(projects_tasks, indent=4))
        # utils.pretty_print(projects_tasks)

        # for i, itemIter in enumerate(projects_tasks):

        #     t = projects_tasks[i]
        #     t['id'] = str(t['id'])

        #     tasks[t['id']] = {
        #         'id':t['id'],
        #         'label':t['title'],
        #         'parent_id':p['id'],
        #         'parent_label':p['title'],
        #         'todolist':user_conf['id'],
        #         'data':t
        #     }

        #     if t['done'] == True:
        #         tasks[t['id']]['status'] = False


        #     # Show favorites in main menu 
        #     if t['is_favorite'] == True:
        #         tasks[t['id']]['priority'] = 1
        #     elif t['priority'] in priorities:
        #         print("Priority task", t['title'])
        #         tasks[t['id']]['priority'] = priorities[t['priority']]
        #         print(tasks[t['id']])

    # beware per_page config ( generally in /opt/vikunja/config.yml) maxitemsperpage defaults to 20 so this really won't work on a default setup 
    #   maxitemsperpage: 5000
    # ?per_page=10000



    response = requests.get(user_conf['url']+"api/v1/tasks/all", headers=headers)
    all_tasks = response.json()

    # print("response.headers")
    # print(response.headers)
    if int(response.headers['X-Pagination-Total-Pages']) > 1:

        # TODO: instead of complaining, iterate through pages 
        utils.error_notice("Vikunja Task Loading Error "," Your vikunja Max items per page setting seems to be truncating your task list! Please increase 'maxitemsperpage' in your vikunja config yml file.")

    # print(json.dumps(all_tasks, indent=4))
    utils.dbg("Got ",len(all_tasks),"Vikunja tasks", s='vikunja', l=2 )

    utils.dbg("Got ",len(all_tasks),"Vikunja tasks", s='vikunja', l=2 )

    for i, itemIter in enumerate(all_tasks):

        t = all_tasks[i]
        id = vikunja_item_id(t['id'], 'task', user_conf)
        list_id = vikunja_item_id(t['project_id'], 'list', user_conf)

        # utils.dbg(t, s='vikunja', l=3)

        # Limit to asks assigned, created by, or favorated by user
        if user_conf["username"]:
            if t['is_favorite'] != True:

                if user_conf['username'] != t['created_by']['username']:

                    if not t['assignees'] or not any(assigned['username'] == user_conf['username'] for assigned in t['assignees']):
                       
                        utils.dbg("skipping "+str(t['title'])+" because it's not assigned to, created by, or a fave of the user",'created_by', t['created_by']['username'],'assignees',t['assignees'], s='vikunja', l=3)
                        continue
        else:
            utils.dbg('no username in vikunja user conf, can\'t limit to relevant tasks', s='vikunja', l=3)


        tasks[id] = {
            'id':id,
            'label':t['title'],
            'parent_id':list_id,
            'parent_label':'',
            'todolist':user_conf['id'],
            'status':1,
            'data':t
        }

        try:
            tasks[id]['parent_label'] = lists[list_id]['label']

            # If lists is inactive make task inactive (This is never the case AFAIK)
            if not lists[list_id]['status']:
                tasks[id]['status'] = 0            

            # If lists is archived make task deleted 
            if lists[list_id]['status'] == -1:
                tasks[id]['status'] = -1

        except Exception as e:
            print("No list found for", t['title']," list_id",list_id)
            # raise e

        if t['done'] == True:
            tasks[id]['status'] = 0

        # Prioritize if "due_date" is soon or passed (and not: "0001-01-01T00:00:00Z")
        if 'due_date' in t and t['due_date'] != "0001-01-01T00:00:00Z":

            due_date = datetime.strptime(t['due_date'],'%Y-%m-%dT%H:%M:%SZ')
            # print(t['title'],"task has a due_date", str(t['due_date']) )
            # print('time_difference', utils.time_difference(due_date,datetime.now()))
            due_seconds = utils.time_difference(due_date,datetime.now())
            # due in next 48 hrs (2 * 24 * 60 * 60) = 172800
            # due today (24 * 60 * 60) = 86400
            if due_seconds > -86400:
                tasks[id]['priority'] = 2

        # prioritize favorites
        if t['is_favorite'] == True:
            tasks[id]['priority'] = 2
        elif t['priority'] in priorities:
            print("Priority task", t['title'])
            tasks[id]['priority'] = priorities[t['priority']]
            # print(tasks[id])

    todos = {'lists':lists,'tasks':tasks}
    return todos


def open(user_conf, item = None, category = None):
    ''' Open Vikunja '''
    url = user_conf['url']

    if item and category:
        if category == 'list':
            url += 'projects/'+str(item['data']['id'])
        elif category == 'task':
            url += 'tasks/'+str(item['data']['id'])

    utils.open_external(url)


import os
import conf
import utils 

def add_new_task(user_conf,list,task_label):
    utils.dbg('Add_task list',list, 'task_label',task_label,s='txt',l=2)
    file_uri = user_conf['file']


    with open(file_uri, 'r') as file:
        # read a list of lines into data
        data = file.readlines()

        if list['id'] == user_conf['id']:
            # Top level insert
            new_line = task_label+"\n"
            line_no = len(data) + 1
        else:
            # Sub list insert
            # TODO: Add it at the end of the list rather than the start ...

            indent = len(list['data']['original_line']) - len(list['data']['original_line'].lstrip()) + 4
            
            new_line = task_label+"\n"
            # I'm sure there's a better way to add a bunch of spaces to to a string
            while indent > 0:
                new_line = " " + new_line
                indent = indent - 1

            line_no = get_line_no(list,data) + 1

        data.insert(line_no,new_line)
        
        # write everything back
        with open(file_uri, 'w') as file:
            file.writelines( data )

        t = {
            # 'id':,
            'label':task_label,
            'parent_id':list['id'],
            'parent_label':list['label'],
            'todolist':user_conf['id'],
            'data':{
                'line_no':line_no,
                'original_line':new_line
            }
        }
        t['id'] = task_id(t)

        if task_label[0].isdigit():
            t['priority'] = float(task_label.split()[0])
        print('new task',t)
        # now the line_no for everything after this will be screwed up...

    return t
            


def get_line_no(i,lines):
    if lines[i['data']['line_no']] == i['data']['original_line']:
        return i['data']['line_no']
    else:

        line_no = 0 #TODO: Improve, (better foreach with key)
        in_list = False
        
        for line in lines:

            # TODO: Find the list 
            if line == i['data']['original_line']:
                return line_no

            line_no = line_no + 1

    throw("Line not found error Ha!")

def task_id(task):
    ''' The task parameter must have at least "label", "todolist" and "parent_id" '''
    # file:sublist/sublist/sublist:label

    id = conf.user['todolists'][task['todolist']]['file']+":"+"/".join(utils.get_lists_for_item(task)['labels'])+":"+task['label']
    print('generated task id',id )
    return id


def mark_task_done(task):
    
    file_uri = conf.user['todolists'][task['todolist']]['file']
    line_no = task['data']['line_no']

    with open(file_uri, 'r') as file:
        # read a list of lines into data
        data = file.readlines()

        print(file_uri +' before mark done:\n')
        print(data)

        if data[line_no] == task['data']['original_line']:
            if data[line_no].strip().startswith('[ ]'):
                data[line_no] = data[line_no].replace('[ ]', '[x]')
            else:
                data[line_no] = data[line_no].replace(task['label'], '[x] '+ task['label'] )

            # write everything back
            with open(file_uri, 'w') as file:
                file.writelines( data )
            return True

        else: 
            print("original line does not match current file line. perhaps it's been edited?")
            return False


def get_todos(user_conf):
    
    file_uri = user_conf['file']

    tasks = {}
    lists = {}
    lists[user_conf['id']] = {
                    'id':user_conf['id'],
                    'label':user_conf['label'],
                    'parent_id':'',
                    'parent_label':'',
                    'todolist':user_conf['id'],
                    'data':{
                        'line_no':0,
                        'original_line':'',
                        'accepts_tasks':True
                    }
                }

    current_list = []
    current_list_label = []
    current_indent = []
    current_indent.append(0)

    parent = ''
    prev_indent = 0
    prev_label = ''
    prev_id = ''

    # id syntax 
    # file:sublist/sublist/sublist:label

    #for testing
    # file_uri = "todo.txt"
    line_no = -1

    with open(file_uri) as file:
        for line in file:

            line_no = line_no + 1

            # Skip blank lines 
            if line.strip() == "":
                continue
            
            label = line.strip()

            # Handle done tasks
            status = 1

            if label.startswith('[x]') or label.startswith('[X]'):
                label = label.removeprefix('[x]').removeprefix('[X]').strip()
                
                status = 0
                print(label, 'set status to ', status)


            label = label.removeprefix('[ ]').removeprefix('[]').strip()
            #TODO: use markdown title syntax as list name in addition to indentation 

            indent = len(line) - len(line.lstrip())

            if indent > prev_indent:
                # prev item is a sub_list header, move it to lists
                lists[prev_id] = tasks[prev_id]
                del tasks[prev_id]
                current_list.append(prev_id)
                current_list_label.append(prev_label)
                current_indent.append(indent)

                utils.dbg('Entered sublist: '+"/".join(current_list_label),s='txt',l=3)

            else: 
                if indent < prev_indent:
                
                    while indent < current_indent[-1]:
                        utils.dbg('Exit sublist: '+"/".join(current_list_label),s='txt',l=3)
                        if current_list:
                            del current_list[-1]
                            del current_list_label[-1]
                            del current_indent[-1]

                    
            utils.dbg(line.rstrip(),s='txt',l=3)
            
            task_id = file_uri+":"+"/".join(current_list_label)+":"+label

            if current_list:
                parent_id = current_list[-1]
                parent_label = current_list_label[-1]
            else:
                parent_id = user_conf['id']
                parent_label = user_conf['label']

            tasks[task_id] = {
                    'id':str(task_id),
                    'label':str(label),
                    'parent_id':str(parent_id),
                    'parent_label':str(parent_label),
                    'todolist':user_conf['id'],
                    'status': status,
                    'data':{
                        'line_no':line_no,
                        'original_line':line,
                        'accepts_tasks':True
                    }
                }

            # Get prioritized tasks (starting with a number)
            # if label[0].isdigit()
            if label.split()[0].isnumeric():
                tasks[task_id]['priority'] = float(label.split()[0])
                tasks[task_id]['label'] = label.removeprefix(label.split()[0]).strip()

            prev_indent = indent
            prev_label = label
            prev_id = task_id

    todos = {'lists':lists,'tasks':tasks}  
    utils.dbg('todos',todos,s='txt',l=3)
    return todos

# testing
# print(get_todos())


import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, Gdk
import conf
import utils
from utils import *

class TaskWindow(Gtk.Window):
    css_provider = Gtk.CssProvider()
    css_provider.load_from_path("styles.css")
    style_context = Gtk.StyleContext()
    style_context.add_provider_for_screen(
        Gdk.Screen.get_default(), css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
    )
    _instance = None

    def __init__(self,app,passed_data=None):

        if TaskWindow._instance:
        
            if TaskWindow._instance.taskEntry.get_text():
                # Present existing window if there is anything in task entry
                TaskWindow._instance.present()
                self = TaskWindow._instance
                self.show_all()
                return None

                
            TaskWindow._instance.destroy() # get rid of old window instance 
            # TaskWindow._instance.present()
            # return None

        TaskWindow._instance = self

        Gtk.Window.__init__(self, title="What Am I Doing?")

        self.app = app
        session = app.session
        self.set_name("TaskWindow") # set css id

        self.fullscreen()
        self.get_style_context().add_class("large")

        # self.present() # try to focus the window
        # self.focus_force() # try to focus the window

        self.shown_tasks = {}
        
        self.set_border_width(30)
        self.set_position(position=1) # Works on x11 but not wayland (ubuntu 22.04)

        accel_group = Gtk.AccelGroup()
        self.add_accel_group(accel_group)

        box = Gtk.VBox(spacing=20)
        box.set_halign(Gtk.Align.CENTER)
        # box.set_valign(Gtk.Align.CENTER)
        
        self.add(box)

        self.session_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=15, border_width=10)

        # self.session_box.set_halign(Gtk.Align.CENTER)
        # box.add(self.session_box)
        box.pack_start(self.session_box,False, False, 0)


        # if self.app.is_running == "boo": #For testing
        if self.app.is_running == True: 

            self.session_label = Gtk.Button(label=session['label']+" "+sec_to_time(session['duration']) )
            # self.session_label.set_relief(Gtk.RELIEF_NONE)
            self.session_label.set_relief(Gtk.ReliefStyle.NONE)
            self.session_label.connect("clicked", self.app.open_session_options_dialog)
            self.session_label.set_property("tooltip-text", "Edit session")

            self.session_box.pack_start(self.session_label, False, False, 0)

            self.session_box.pack_start(Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=15, border_width=10),True, True, 0)

            pause_button = Gtk.Button()
            # pause_button.set_image(Gtk.Image.new_from_file(os.path.abspath('icon/pause.svg'))) 
            pause_button.set_image(Gtk.Image.new_from_file(os.path.abspath('icon/pause.png')))
            pause_button.connect("clicked", self.app.stop_task)
            pause_button.connect("clicked", self.update_default_tasks) 
            pause_button.set_property("tooltip-text", "Pause Task (Ctrl + P)")
            pause_button.set_relief(Gtk.ReliefStyle.NONE)
            self.session_box.add(pause_button)

            done_button = Gtk.Button()
            # done_button.set_image(Gtk.Image.new_from_file(os.path.abspath('icon/mark-done.svg')))
            done_button.set_image(Gtk.Image.new_from_file(os.path.abspath('icon/mark-done.png')))
            done_button.set_property("tooltip-text", "Mark Task Done (Ctrl + D)")
            done_button.connect("clicked", self.app.stop_task,'mark_done')
            done_button.connect("clicked", self.update_default_tasks) 
            done_button.set_relief(Gtk.ReliefStyle.NONE)
            self.session_box.add(done_button)


            cancel_button = Gtk.Button()
            # cancel_button.set_image(Gtk.Image.new_from_file(os.path.abspath('icon/cancel.svg'))) 
            cancel_button.set_image(Gtk.Image.new_from_file(os.path.abspath('icon/cancel.png'))) 
            cancel_button.connect("clicked", self.app.stop_task,"cancel")
            cancel_button.set_property("tooltip-text", "Discard timer (Ctrl + X)")
            cancel_button.set_relief(Gtk.ReliefStyle.NONE)
            self.session_box.add(cancel_button)



            # self.notes_text_buffer = Gtk.TextBuffer()
            # if 'notes' in self.session:
            #     self.notes_text_buffer.set_text(session['notes'])
            # else:
            #     self.notes_text_buffer.set_text('\n\n') # hack to give it some height

            # self.notes = Gtk.TextView(buffer=self.notes_text_buffer)
            # box.add(self.notes)
                

            self.tick_timer = GLib.timeout_add_seconds(1, self.tick)

        else:
            # RelevantQuestion prompt
            print("show the question")
            # self.session_box.foreach(lambda child: child.destroy()) # for testing

            lines = conf.user['prompts'].split("\n")
            p = lines.pop(0)
            lines.append(p)
            conf.user['prompts'] = "\n".join(lines)
            self.the_question = Gtk.Label(p)
            self.the_question.set_name("RelevantQuestion")
            self.session_box.pack_start(self.the_question, True, True, 0)

                            
        # Large fuzzy task input 
        self.taskEntry = Gtk.Entry()
        self.taskEntry.set_name("FuzzyTask") # set css id
        self.taskEntry.set_width_chars(59)
        self.taskEntry.set_placeholder_text("Find Task")
        
        # box.add(self.taskEntry)
        box.pack_start(self.taskEntry,False, False, 0)

        self.taskEntry.grab_focus()

        if passed_data:
            dbg('taskwindow passed_data',passed_data,s='taskwindow')
            if 'afk_time' in passed_data:

                last_active_time = datetime.now() - timedelta(seconds=passed_data['afk_time'])
                # utc_last_active_time = datetime.now(timezone.utc) - timedelta(seconds=passed_data['afk_time'])

                last_active_str = time.strftime('%H:%M', last_active_time.timetuple())

                afk_label = Gtk.Label()
                afk_label.set_markup("<b>Inactive Since "+" "+str(last_active_str) +"</b>")
                self.session_box.add(afk_label)

                pause_then_button = Gtk.Button(label="Finish Then")

                pause_then_button.connect("clicked", self.pause_then, last_active_time)
                self.session_box.add(pause_then_button)

            if 'task' in passed_data:
                self.taskEntry.set_text(passed_data['task']['label'])
        

        # box.add(Gtk.Box(border_width=10)) # Spacer

        self.scrolled_window = Gtk.ScrolledWindow()
        self.scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        box.pack_start(self.scrolled_window, True, True, 0)
        # self.scrolled_window.set_size_request(-1, -1)
        self.scrolled_window.set_size_request(-1, 400)
        # self.scrolled_window.set_hexpand(True)
        # self.scrolled_window.set_vexpand(True)
        
        self.tasks_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=7)
        self.tasks_box.set_halign(Gtk.Align.START)

        self.scrolled_window.add(self.tasks_box)

        self.total_duration_label = Gtk.Label()
        box.pack_start(self.total_duration_label,False, False, 0)


        self.buttons_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=20)
        self.buttons_box.set_halign(Gtk.Align.CENTER)

        # box.add(Gtk.Box(border_width=10)) # Spacer

        # box.add(self.buttons_box)
        box.pack_start(self.buttons_box,False, False, 0)

        self.settings_button = Gtk.Button(label="Settings")
        self.settings_button.connect("clicked",self.app.open_settings_window)
        self.buttons_box.add(self.settings_button)

        self.refresh_button = Gtk.Button(label="Refresh")
        self.refresh_button.connect("clicked",self.app.async_refresh)
        self.buttons_box.add(self.refresh_button)

        # Todolist openables
        openables = []
        for id, todolist in conf.user['todolists'].items():
            if todolist['status']:
                openable = get_connector_openable(None, todolist,False)
                if openable not in openables:            
                    openables.append(openable)
                    
                    openable_button = Gtk.Button(label=todolist['label'])
                    openable_button.connect("clicked", get_connector_openable, todolist)
                    self.buttons_box.add(openable_button)
            

        self.new_task_button = Gtk.Button(label="New Task")
        self.new_task_button.connect("clicked",self.open_new_task_dialog)
        self.buttons_box.add(self.new_task_button)


        key, mod = Gtk.accelerator_parse('<Control>n')
        self.new_task_button.add_accelerator("clicked", accel_group, key, mod, Gtk.AccelFlags.VISIBLE)

        key, mod = Gtk.accelerator_parse('<Control>r')
        self.refresh_button.add_accelerator("clicked", accel_group, key, mod, Gtk.AccelFlags.VISIBLE)
        # dbg('new_task_button accel','key',key, 'mod',mod,s='taskwindow',l=2)

        if self.app.is_running == True: 

            key, mod = Gtk.accelerator_parse('<Control>p')
            pause_button.add_accelerator("clicked", accel_group, key, mod, Gtk.AccelFlags.VISIBLE)

            key, mod = Gtk.accelerator_parse('<Control>x')
            cancel_button.add_accelerator("clicked", accel_group, key, mod, Gtk.AccelFlags.VISIBLE)

            key, mod = Gtk.accelerator_parse('<Control>d')
            cancel_button.add_accelerator("clicked", accel_group, key, mod, Gtk.AccelFlags.VISIBLE)


        self.connect("key-press-event", self.on_win_key_press_event)
        self.connect("window-state-event", self.on_window_state_event)

        self.taskEntry.connect("changed",self.task_search)
        
        self.update_default_tasks()
        # dbg('recent_tasks',recent_tasks,s='taskwindow', l=3)
        # dbg('self.default_tasks',self.default_tasks,s='taskwindow')

        self.show_all()


    def update_default_tasks(self,w = None):
        self.default_tasks = utils.get_priority_tasks(50) | utils.get_recent_tasks(30)
        self.task_search(self.taskEntry)
        

    def open_new_task_dialog(self,w = None):
        passed_data = {"label":self.taskEntry.get_text()}
        self.app.open_new_task_dialog(self, passed_data)
    

    def tick(self):

        if(self.app.is_running == True):
            self.session_label.set_label(self.app.session['label'] + ": " + sec_to_time(self.app.session['duration']))
            self.session_label.show()
            return True
        else:
            self.session_box.hide()
            return False


    def pause_then(self, widget = None, utc_last_active_time = None):
        self.app.stop_task(None,"save",utc_last_active_time)
        self.session_box.destroy()
        self.update_default_tasks()


    def task_search(self,widget):

        self.tasks_box.foreach(lambda child: child.destroy()) 
        self.shown_tasks.clear()

        # TODO: cache this and avoid duplicate searches
        i = widget.get_text()

        # TODO: add "default time tracking range" to settings then add " WHERE sessions.start_time > conf.get_default_tracking_range_as_time('sql')" 

        utils.dbg({"task search":i},s='taskwindow')
        if len(i) == 0:
            tasks = self.default_tasks
            # self.new_task_button.hide()

        else:
            if  i == '*':
                # data = utils.db_query("SELECT tasks.*, SUM(sessions.duration) as duration FROM tasks JOIN sessions ON sessions.task_id = tasks.id  WHERE tasks.status = '1' ORDER BY status DESC, priority DESC, extended_label ASC LIMIT 5000")
                data = utils.db_query("SELECT tasks.*, SUM(sessions.duration) as duration FROM tasks LEFT JOIN sessions ON sessions.task_id = tasks.id WHERE tasks.status IS NOT '-1' GROUP BY tasks.id ORDER BY duration DESC LIMIT 5000")

            elif len(i) < 3:
                #  look ahead, if less than 3 chars
                data = utils.db_query("SELECT * FROM tasks WHERE (label like ? OR parent_label like ?) AND status IS NOT '-1' ORDER BY status DESC, priority DESC, extended_label ASC LIMIT 100",(i+'%',i+'%',))

                # NOTE: Unfortunately, can't use "AND status >= 0" because, for some reason, some (Most?) completed asks are status NULL, also status != '-1' omits NULL (completed) row (because NULL is never 'equal' OR 'not equal' to anything)
            else:

                data = utils.db_query("SELECT tasks.*, SUM(sessions.duration) as duration FROM tasks LEFT JOIN sessions ON sessions.task_id = tasks.id WHERE tasks.extended_label like ? AND tasks.status IS NOT '-1' GROUP BY tasks.id ORDER BY tasks.status DESC, tasks.priority DESC, tasks.extended_label ASC LIMIT 500",('%'+i+'%',))

            tasks = {}

            for t in data:
                tasks[t['id']] = proc_db_item(t)

            # self.new_task_button.show()

        utils.dbg('task_window search tasks',tasks,s='taskwindow',l=3)

        total_duration = 0

        for id, t in tasks.items():
            self.add_task_to_list(t,GLib.markup_escape_text(i))
            if 'duration' in t and t['duration']:
                total_duration += int(t['duration'])

        if total_duration:
            self.total_duration_label.set_markup('<b>'+str(round(total_duration / 60 / 60,2))+'</b> hours')
        else:
            self.total_duration_label.set_markup('')

        self.tasks_box.show_all()


    def add_task_to_list(self,t,search_str = None):

        try:
            utils.dbg("add_task_to_list "+ t['label'], "status",t['status'], s='taskwindow',l=3)

            self.shown_tasks[t['id']] = Gtk.Button()
            self.shown_tasks[t['id']].set_halign(Gtk.Align.START)
            self.shown_tasks[t['id']].set_hexpand(True)

            label = Gtk.Label()

            # self.shown_tasks[t['id']].set_label(label)
            # button_context = self.shown_tasks[t['id']].get_style_context().add_class("large")
            extended_label = GLib.markup_escape_text(t['extended_label'])
            
            if search_str:
                extended_label = extended_label.replace(search_str,"<b>"+search_str+"</b>")
                extended_label = extended_label.replace(search_str.capitalize(),"<b>"+search_str.capitalize()+"</b>") # Cheesy
            if "duration" in t and t['duration']:
                extended_label += " ("+sec_to_time(t['duration'])+")"

            if not t['status']:
                utils.dbg("add strikethrough to done task "+t['label'],l=3,s="taskwindow")
                # button_context.add_class("done")
                label.set_markup('<s>'+extended_label+'</s>')

            elif t['label'] == self.taskEntry.get_text() or t['priority'] > 0:
                # button_context.add_class("bold")
                label.set_markup('<b>'+extended_label+'</b>')
            else:
                label.set_markup(extended_label)

            self.shown_tasks[t['id']].add(label)

            # self.shown_tasks[t['id']].set_size_request(955, -1)
            self.shown_tasks[t['id']].connect("clicked", self.select_task,None,t)
            self.shown_tasks[t['id']].connect("button-release-event", self.select_task,t)
            self.shown_tasks[t['id']].set_relief(Gtk.ReliefStyle.NONE)
            self.tasks_box.add(self.shown_tasks[t['id']])

        except Exception as e:
            utils.dbg("Error adding task to list"+ t['label'], e, l=0, s='taskwindow')


    def select_task(self,widget,event=None,t=None):
        print('event',event)
        if event:  # Right-click (button 3)
            if event.button == 3:  # Right-click (button 3)
                # error_notice("Thank you for right clicking!", "Watch this space for exciting new features!")
                # TODO: Context menu with: Modify session to this task, mark_done, start, edit?, and get_times? (from time tracker)

                try:
                    c = conf.user['todolists'][t['todolist']]
                    conf.todo_connectors[c['type']].open(c,t,'task')
                    return
                except Exception as e:
                    error_notice('Bonk', "error with "+ c['type']+ " open function ")


        self.app.start_task(None,t) 
        self.destroy()


    def fullscreen_mode(self):

        if self.__is_fullscreen:
            self.unfullscreen()
            self.get_style_context().remove_class("large")

        else:
            self.fullscreen()
            self.get_style_context().add_class("large")


    def on_win_key_press_event(self, widget, ev):

        key = Gdk.keyval_name(ev.keyval)
        utils.dbg("task window key",key,s="taskwindow",l=2)
        # utils.dbg("key_press_event",ev,s="taskwindow",l=3)

        if key == "F11":
            self.fullscreen_mode()
        elif key == "Escape":
            self.destroy()
        elif key == "Return" and self.taskEntry.is_focus():
            # print("self.taskEntry has focus, do the thing")
            if self.shown_tasks:
                self.tasks_box.get_children()[0].emit('clicked')
            else:
                self.new_task_button.emit('clicked')

    def on_window_state_event(self, widget, ev):
        self.__is_fullscreen = bool(ev.new_window_state & Gdk.WindowState.FULLSCREEN)




import os
import conf
import utils 
from taskw import TaskWarrior

# lists and tasks are each a dicts of dicts with following shape 
# tasks[task_id] = {
#     'id':task_id,
#     'label':label,
#     'parent_id':parent_id,
#     'parent_label':parent_label,
#     'todolist':user_conf['id'],
#     'data':{
#         # Any additional data
#     }
# }

def add_new_task(user_conf,list,task_label):
    ''' Adds the label to the provided list in the specified todo and returns a task dict'''
    w = TaskWarrior()
    twt = w.task_add(task_label, project=list['label'])

    # {'id': 2, 'description': 'TW add task W SCP', 'entry': '20241024T013615Z', 'modified': '20241024T013615Z', 'project': 'SCP', 'status': 'pending', 'uuid': '240f964e-e459-4f04-8a28-cbf61f44a82a', 'urgency': 1}

    t = {
        'id':twt['uuid'],
        'label':twt['description'],
        'parent_id':user_conf['id'],
        'parent_label':user_conf['label'],
        'todolist':user_conf['id'],
        'data':twt
    }

    if 'project' in twt:
        t['parent_id'] = twt['project']
        t['parent_label'] = twt['project']

    if 'priority' in twt:
        if twt['priority'] == 'H':
            t['priority'] = 1

        if twt['priority'] == 'M':
            t['priority'] = 3
    dbg('add_new_task result',t,s='TaskWarrior')
    return t
            

def mark_task_done(task):
    '''Return True on success False on error'''
    w = TaskWarrior()
    w.task_done(id=task['data']['id'])


def get_todos(user_conf):
    '''Return {'lists':lists,'tasks':tasks} '''
    
    tasks = {}
    lists = {
        user_conf['id']: {
            'id':user_conf['id'],
            'label':user_conf['label'],
            'parent_id':'',
            'parent_label':'',
            'todolist':user_conf['id'],
            'status': 1,
            'data':{
                'accepts_tasks':True
            }
        }
    }

    
    w = TaskWarrior()
    tw_tasks = w.load_tasks()

    # TODO: also add tasks from tw_tasks['completed'] with status 0

    for twt in tw_tasks['pending']:
        t = {
            'id':twt['uuid'],
            'label':twt['description'],
            'parent_id':user_conf['id'],
            'parent_label':user_conf['label'],
            'todolist':user_conf['id'],
            'data':twt
        }

        tasks[twt['uuid']] = t

        if 'project' in twt:

            t['parent_id'] = twt['project']
            t['parent_label'] = twt['project']

            if twt['project'] not in lists:
                lists[twt['project']] = {
                    'id':twt['project'],
                    'label':twt['project'],
                    'parent_id':user_conf['id'],
                    'parent_label':user_conf['label'],
                    'todolist':user_conf['id'],
                    'data':{
                        'accepts_tasks': True
                    }
                }

        if 'priority' in twt:
            if twt['priority'] == 'H':
                t['priority'] = 5

            if twt['priority'] == 'M':
                t['priority'] = 2


    todos =  {'lists':lists,'tasks':tasks}
    return todos

# testing
# print(get_todos())


import os
import conf
import utils 

# lists and tasks are each a dict of dicts with following shape 
# tasks[task_id] = {
#     'id':task_id,
#     'label':label,
#     'parent_id':parent_id,
#     'parent_label':parent_label,
#     'status':1, #( 1 pending, 0 done, -1 deleted)
#     'priority':1 max, 5 min, 0 none,
#     'todolist':user_conf['id'],
#     'data':{
#         # Any additional data

#     }
# }

def add_new_task(user_conf,list,task_label):
    ''' Adds the label to the provided list in the specified todo and returns a task dict'''

    return t
            
def mark_task_done(task):
    '''Return True on success False on error'''

def get_todos(user_conf):
    '''Return {'lists':lists,'tasks':tasks} '''
    
    # todo_conf is conf.user['todolists'][user_conf['id']]
    tasks = {}
    lists = {
        user_conf['id']: {
            'id':user_conf['id'],
            'label':user_conf['label'],
            'parent_id':'',
            'parent_label':'',
            'todolist':user_conf['id'],
            'status': 1,
            'data':{
                'accepts_tasks':False
            }
        }
    }


    todos =  {'lists':lists,'tasks':tasks}
    return todos


def open(user_conf, item = None, category = None):
    ''' Open todolist '''
    url = user_conf['url']

    if item and category:
        if category == 'list':
            url += ''
        elif category == 'task':
            url += ''

    utils.open_external(url)

# testing
# print(get_todos())